from typing import Any, Self

from boa3_test.test_drive.model.network.payloads.transactionattribute import TransactionAttribute, \
    TransactionAttributeType


class OracleResponse(TransactionAttribute):
    def __init__(self, id_: int, code: str, result: str):
        super().__init__(TransactionAttributeType.ORACLE_RESPONSE)
        self._id = id_
        self._code = code
        self._result = result

    def to_json(self) -> dict[str, Any]:
        import base64
        import json
        json_response = super().to_json()

        json_response['id'] = self._id
        json_response['code'] = self._code
        json_response['result'] = json.loads(base64.b64decode(self._result))

        return json_response

    @classmethod
    def from_json(cls, json: dict[str, Any]) -> Self:
        oracle_response: OracleResponse = OracleResponse(json['id'], json['code'], json['result'])

        return oracle_response



import ast

from boa3.internal.model.callable import Callable
from boa3.internal.model.debuginstruction import DebugInstruction
from boa3.internal.model.symbol import ISymbol
from boa3.internal.model.type.classes.classtype import ClassType
from boa3.internal.model.type.type import IType, Type
from boa3.internal.model.variable import Variable


class Method(Callable):
    """
    A class used to represent a function or a class method

    :ivar args: a dictionary that maps each arg with its name. Empty by default.
    :ivar locals: a dictionary that maps each local variable with its name. Empty by default.
    :ivar imported_symbols: a dictionary that maps each imported symbol with its name. Empty by default.
    :ivar is_public: a boolean value that specifies if the method is public. False by default.
    :ivar return_type: the return type of the method. None by default.
    """

    def __init__(self,
                 args: dict[str, Variable] = None,
                 vararg: tuple[str, Variable] | None = None,
                 kwargs: dict[str, Variable] | None = None,
                 defaults: list[ast.AST] = None,
                 return_type: IType = Type.none, is_public: bool = False,
                 decorators: list[Callable] = None,
                 is_init: bool = False,
                 external_name: str = None,
                 is_safe: bool = False,
                 origin_node: ast.AST | None = None,
                 deprecated: bool = False
                 ):
        super().__init__(
            args=args,
            vararg=vararg,
            kwargs=kwargs,
            defaults=defaults,
            return_type=return_type,
            is_public=is_public,
            decorators=decorators,
            external_name=external_name,
            is_safe=is_safe,
            origin_node=origin_node,
            deprecated=deprecated
        )

        self.imported_symbols = {}
        self._symbols = {}
        self.defined_by_entry = True
        self.is_init = is_init
        self.locals: dict[str, Variable] = {}

        if is_init and self.has_cls_or_self:
            self.return_type = list(self.args.values())[0].type

        self._debug_map: list[DebugInstruction] = []
        self.origin_class: ClassType | None = None
        self.file_origin: str | None = None

    @property
    def shadowing_name(self) -> str:
        return 'method'

    def include_variable(self, var_id: str, var: Variable, is_global: bool = False):
        """
        Includes a variable into the list of locals

        :param var_id: variable identifier
        :param var: variable to be included
        :param is_global: whether the variable is declared outside the method
        """
        if var_id not in self.symbols:
            if is_global:
                self._symbols[var_id] = var
            else:
                self.locals[var_id] = var

    @property
    def symbols(self) -> dict[str, Variable]:
        """
        Gets all the symbols in the method

        :return: a dictionary that maps each symbol in the module with its name
        """
        symbols = self._symbols.copy()
        symbols.update(self.imported_symbols)
        symbols.update(self.args)
        symbols.update(self.locals)
        return symbols

    def include_symbol(self, symbol_id: str, symbol: ISymbol, is_global: bool = False):
        """
        Includes a method into the scope of the module

        :param symbol_id: method identifier
        :param symbol: method to be included
        :param is_global: whether the variable is declared outside the method
        """
        if symbol_id not in self.symbols:
            if isinstance(symbol, Variable):
                self.include_variable(symbol_id, symbol, is_global)
            else:
                self.imported_symbols[symbol_id] = symbol

    @property
    def origin(self) -> ast.AST:
        """
        Returns the method origin ast node.

        :return: the ast node that describes this method. None if it is not from a ast.
        """
        return self._origin_node

    def debug_map(self) -> list[DebugInstruction]:
        """
        Returns a list with the debug information of each mapped Python instruction inside this method
        """
        from boa3.internal.compiler.codegenerator.vmcodemapping import VMCodeMapping
        return sorted(self._debug_map, key=lambda instr: VMCodeMapping.instance().get_start_address(instr.code))

    def include_instruction(self, instr_info: DebugInstruction):
        """
        Includes a new instruction in the debug info

        :param instr_info: debug information from the new instruction
        """
        if not any((info.start_line == instr_info.start_line and info.start_col == instr_info.start_col
                    for info in self._debug_map)):
            existing_instr_info: DebugInstruction | None = \
                next((info for info in self._debug_map if info.code == instr_info.code), None)
            if existing_instr_info is not None:
                self._debug_map.remove(existing_instr_info)
            self._debug_map.append(instr_info)

    def remove_instruction(self, start_line: int, start_col: int):
        """
        Removes a instruction from the debug info at the given position if it exists

        :param start_line: instruction's first line
        :param start_col: instruction's beginning offset in the first line
        """
        instruction = next((info for info in self._debug_map
                            if info.start_line == start_line and info.start_col == start_col),
                           None)
        if instruction is not None:
            self._debug_map.remove(instruction)

    def args_to_be_generated(self) -> list[int]:
        """
        Gets the indexes of the arguments that must be generated.
        If method has `self` arg, it must be in this list.

        By default, all the arguments are generated.

        :return: A list with the indexes of the arguments that must be generated
        """
        return list(range(len(self.args)))



from boa3.internal.model.builtin.contract.nep17_interface_methods.nep17interfacemethod import Nep17InterfaceMethod
from boa3.internal.model.builtin.interop.contract import ContractType
from boa3.internal.model.variable import Variable


class Nep17InterfaceTotalSupplyMethod(Nep17InterfaceMethod):

    def __init__(self, self_type: ContractType):
        from boa3.internal.model.type.type import Type

        args: dict[str, Variable] = {'self': Variable(self_type)}
        super().__init__(args, 'total_supply', native_identifier='totalSupply', return_type=Type.int)



from enum import IntEnum


class VMState(IntEnum):
    """
    Represents the VM execution state.
    """
    NONE = 0
    """
    Indicates that the execution is in progress or has not yet begun.

    :meta hide-value:
    """

    HALT = 1
    """
    Indicates that the execution has been completed successfully.

    :meta hide-value:
    """

    FAULT = 2
    """
    Indicates that the execution has ended, and an exception that cannot be caught is thrown.

    :meta hide-value:
    """

    BREAK = 4
    """
    Indicates that a breakpoint is currently being hit.

    :meta hide-value:
    """

    def __str__(self) -> str:
        return self.name


def get_vm_state(state_name: str) -> VMState:
    try:
        return VMState[state_name]
    except BaseException:
        return VMState.NONE



import json
import logging
import os.path
from typing import Any

from boa3.internal import constants
from boa3.internal.analyser.analyser import Analyser
from boa3.internal.compiler.compileroutput import CompilerOutput
from boa3.internal.compiler.filegenerator.importdata import ImportData
from boa3.internal.model.event import Event
from boa3.internal.model.imports.importsymbol import BuiltinImport, Import
from boa3.internal.model.imports.package import Package
from boa3.internal.model.method import Method
from boa3.internal.model.symbol import ISymbol
from boa3.internal.model.type.classes.classtype import ClassType
from boa3.internal.model.type.classes.userclass import UserClass
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo import to_hex_str
from boa3.internal.neo.contracts.neffile import NefFile


class FileGenerator:
    """
    This class is responsible for generating the files.
    """

    def __init__(self, compiler_result: CompilerOutput, analyser: Analyser, entry_file: str):
        import os
        self._metadata = analyser.metadata
        self._symbols: dict[str, ISymbol] = analyser.symbol_table.copy()

        self._entry_file = entry_file
        self._entry_file_full_path = analyser.path.replace(os.sep, constants.PATH_SEPARATOR)

        self._files: list[str] = [self._entry_file_full_path]
        self._nef: NefFile = NefFile(compiler_result.bytecode,
                                     source=self._metadata.source,
                                     method_tokens=compiler_result.method_tokens)

        self.__all_imports: list[Import] = None
        self._all_methods: dict[str, Method] = None
        self._all_static_vars: dict[str, Variable] = None

        self._inner_methods = None
        self._inner_events = None

    @property
    def _public_methods(self) -> dict[str, Method]:
        """
        Gets a sublist of the symbols containing all public methods

        :return: a dictionary that maps each public method with its identifier
        """
        return {name: method for name, method in self._methods.items() if method.is_public}

    @property
    def _static_variables(self) -> dict[str, Variable]:
        """
        Gets a sublist of the symbols containing all global variables

        :return: a dictionary that maps each global variable with its identifier
        """
        if self._all_static_vars is None:
            variables: dict[tuple[str, str], Variable] = {}
            imported_symbols: dict[str, Import] = {}

            for name, symbol in self._symbols.items():
                if isinstance(symbol, Variable) and not symbol.is_reassigned:
                    variables[(self._entry_file, name)] = symbol
                elif isinstance(symbol, Import):
                    imported_symbols[symbol.origin] = symbol

            imported_to_map, imports_unique_ids = self._get_imports_unique_ids(imported_symbols,
                                                                               False,
                                                                               list(variables.values())
                                                                               )

            if imports_unique_ids[-1] != self._entry_file:
                # update entry file to be a unique name
                unique_id = imports_unique_ids[-1]
                variables = {(unique_id, method_name): method for (module_id, method_name), method in variables.items()}

            # include all user created methods in the list, even the methods that aren't imported in the entry file
            for index in range(len(imported_to_map)):
                module_full_path, module_import = list(imported_to_map.items())[index]
                module_id = imports_unique_ids[index]

                for name, symbol in module_import.all_symbols.items():
                    need_to_map = (isinstance(symbol, Variable)
                                   and not symbol.is_reassigned
                                   and symbol not in [variables for variables in variables.values()])
                    if need_to_map:
                        variables[(module_id, name)] = symbol

            if len(imports_unique_ids) > 1:
                self._all_static_vars = {f'{unique_id}{constants.VARIABLE_NAME_SEPARATOR}{var_id}': var
                                         for (unique_id, var_id), var in variables.items()}
            else:
                self._all_static_vars = {var_id: var
                                         for (unique_id, var_id), var in variables.items()}

        return self._all_static_vars

    @property
    def _methods(self) -> dict[str, Method]:
        """
        Gets a sublist of the symbols containing all user methods

        :return: a dictionary that maps each method with its identifier
        """
        if self._inner_methods is None:
            from boa3.internal.model.builtin.method import IBuiltinMethod
            all_entry_file_methods = {}
            for name, symbol in self._symbols.items():
                if isinstance(symbol, Method) and symbol.defined_by_entry:
                    all_entry_file_methods[name] = symbol
                elif isinstance(symbol, IBuiltinMethod) and symbol.is_public:
                    all_entry_file_methods[name] = symbol
                elif isinstance(symbol, UserClass):
                    for class_method_name, class_method in symbol.methods.items():
                        if class_method.defined_by_entry:
                            all_entry_file_methods[f'{symbol.identifier}.{class_method_name}'] = class_method

            self._inner_methods = all_entry_file_methods
        return self._inner_methods

    @property
    def _methods_with_imports(self) -> dict[tuple[str, str], Method]:
        if self._all_methods is None:
            from boa3.internal.model.builtin.decorator.builtindecorator import IBuiltinCallable

            methods: dict[tuple[str, str], Method] = {}
            imported_symbols: dict[str, Import] = {symbol.origin: symbol for symbol in self._all_imports}

            for name, symbol in self._symbols.items():
                if isinstance(symbol, Method) and not isinstance(symbol, IBuiltinCallable):
                    if symbol.defined_by_entry:
                        methods[(self._entry_file, name)] = symbol
                elif isinstance(symbol, UserClass):
                    for class_method_name, class_method in symbol.methods.items():
                        if class_method.defined_by_entry and class_method.is_compiled:
                            methods[(self._entry_file, f'{symbol.identifier}.{class_method_name}')] = class_method

                elif isinstance(symbol, Import) and symbol.origin not in imported_symbols:
                    imported_symbols[symbol.origin] = symbol

            imported_to_map, imports_unique_ids = self._get_imports_unique_ids(imported_symbols,
                                                                               True,
                                                                               list(methods.values())
                                                                               )

            if imports_unique_ids[-1] != self._entry_file:
                # update entry file to be a unique name
                unique_id = imports_unique_ids[-1]
                methods = {(unique_id, method_name): method for (module_id, method_name), method in methods.items()}

            # include all user created methods in the list, even the methods that aren't imported in the entry file
            for index in range(len(imported_to_map)):
                module_full_path, module_import = list(imported_to_map.items())[index]
                module_id = imports_unique_ids[index]

                for name, symbol in module_import.all_symbols.items():
                    if (isinstance(symbol, Method)
                            and not isinstance(symbol, IBuiltinCallable)
                            and symbol not in [method for method in methods.values()]):
                        if symbol.file_origin is None and module_import.origin_file in self._files:
                            symbol.file_origin = module_import.origin_file

                        methods[(module_id, name)] = symbol
                    elif isinstance(symbol, UserClass):
                        for class_method_name, class_method in symbol.methods.items():
                            if class_method.file_origin is None and module_import.origin_file in self._files:
                                class_method.file_origin = module_import.origin_file

                            methods[(module_id, f'{symbol.identifier}.{class_method_name}')] = class_method

            self._all_methods = methods
        return self._all_methods

    @property
    def _events(self) -> dict[str, Event]:
        """
        Gets a sublist of the symbols containing all user events

        :return: a dictionary that maps each event with its identifier
        """
        if self._inner_events is None:
            events = set()
            for imported in self._all_imports:
                events.update([event for event in imported.all_symbols.values() if isinstance(event, Event)])

            def check_package_events(package: Package):
                for package_symbol in package.symbols.values():
                    if isinstance(package_symbol, Event):
                        events.add(package_symbol)
                    elif isinstance(package_symbol, Package):
                        check_package_events(package_symbol)

                for inner_package in package.inner_packages.values():
                    check_package_events(inner_package)

            for symbol in self._symbols.values():
                if isinstance(symbol, Event):
                    events.add(symbol)
                elif isinstance(symbol, Package):
                    check_package_events(symbol)

            self._inner_events = {event.name: event for event in events}
        return self._inner_events

    @property
    def _all_imports(self) -> list[Import]:
        if self.__all_imports is None:
            all_imports = [imported for imported in self._symbols.values()
                           if (isinstance(imported, (Import, Package))
                               and not isinstance(imported, BuiltinImport))]
            only_imports = []
            imported_files = []

            index = 0
            while index < len(all_imports):
                imported = all_imports[index]
                index += 1

                if isinstance(imported, Package) and isinstance(imported.origin, Import):
                    all_imports.append(imported.origin)
                    file_origin = imported.origin.origin

                    for symbol in imported.symbols.values():
                        if isinstance(symbol, Method) and symbol.is_compiled and symbol.file_origin is None:
                            symbol.file_origin = file_origin
                        if isinstance(symbol, UserClass):
                            for class_symbol in symbol.symbols.values():
                                if isinstance(class_symbol, Method) and class_symbol.file_origin is None:
                                    class_symbol.file_origin = file_origin

                if isinstance(imported, Import):
                    if imported.origin not in imported_files:
                        only_imports.append(imported)
                        imported_files.append(imported.origin)
                    else:
                        # import already included
                        continue

                inner_symbols = imported.all_symbols if hasattr(imported, 'all_symbols') else imported.symbols
                for inner in inner_symbols.values():
                    if (isinstance(inner, (Import, Package))
                            and not isinstance(inner, BuiltinImport)
                            and inner not in all_imports):
                        all_imports.append(inner)

            self.__all_imports = list(reversed(only_imports))  # first positions are the most inner imports
            # using for instead a generator to keep the result determined
            for import_ in imported_files:
                if os.path.isfile(import_) and import_ not in self._files:
                    self._files.append(import_)

        return self.__all_imports

    def create_folder(self, output_folder: str):
        if os.path.isfile(output_folder):
            output_folder = os.path.abspath(os.path.dirname(output_folder))
        else:
            output_folder = os.path.abspath(output_folder)

        folders_to_generate = []
        while not os.path.exists(output_folder):
            folders_to_generate.append(output_folder)
            output_folder = os.path.dirname(output_folder)

        while len(folders_to_generate) > 0:
            folder = folders_to_generate.pop()
            os.mkdir(folder)

    # region NEF

    @property
    def _nef_hash(self) -> str:
        """
        Gets the string representation of the hash of the nef file

        :return: the hex string representation of the hash
        """
        return to_hex_str(self._nef.script_hash)

    def generate_nef_file(self) -> bytes:
        """
        Generates the .nef file

        :return: the resulting nef file as a byte array
        """
        return self._nef.serialize()

    # endregion

    # region Manifest

    def generate_manifest_file(self) -> bytes:
        """
        Generates the .manifest metadata file

        :return: the resulting manifest as a byte array
        """
        data: dict[str, Any] = self._get_manifest_info()
        json_data: str = json.dumps(data, indent=4)
        return bytes(json_data, constants.ENCODING)

    def _get_manifest_info(self) -> dict[str, Any]:
        """
        Gets the manifest information in a dictionary format

        :return: a dictionary with the manifest information
        """
        return {
            "name": self._get_name(),
            "groups": self._get_groups(),
            "abi": self._get_abi_info(),
            "permissions": self._get_permissions(),
            "trusts": self._metadata.trusts,
            "features": {},
            "supportedstandards": self._metadata.supported_standards,
            "extra": self._get_extras()
        }

    def _get_name(self) -> str:
        """
        Gets the name of the contract, if name wasn't specified it will be the file name.

        :return: the contract name
        """
        return self._metadata.name if self._metadata.name else self._entry_file

    def _get_permissions(self) -> list[dict[str, Any]]:
        """
        Gets the permission information in a dictionary format, if _metadata._permissions is empty, then consider it
        with the import wildcard inside it.

        :return: a dictionary with the permission information
        """
        return self._metadata.permissions

    def _get_groups(self) -> list[dict[str, Any]]:
        """
        Gets the group information in a dictionary format.

        :return: a dictionary with the groups information
        """
        return self._metadata.groups

    def _get_abi_info(self) -> dict[str, Any]:
        """
        Gets the abi information in a dictionary format

        :return: a dictionary with the abi information
        """
        return {
            "methods": self._get_abi_methods(),
            "events": self._get_abi_events()
        }

    def _get_abi_methods(self) -> list[dict[str, Any]]:
        """
        Gets the abi methods in a dictionary format

        :return: a dictionary with the abi methods
        """
        return [
            self._construct_abi_method(method_id, method)
            for method_id, method in self._public_methods.items()
        ]

    def _construct_abi_method(self, method_id: str, method: Method) -> dict[str, Any]:
        from boa3.internal.compiler.codegenerator.vmcodemapping import VMCodeMapping

        abi_method_name = method.external_name if isinstance(method.external_name, str) else method_id
        logging.getLogger(constants.BOA_LOGGING_NAME).info(f"'{abi_method_name}' method included in the manifest")

        method_abi = {
            "name": abi_method_name,
            "offset": (VMCodeMapping.instance().get_start_address(method.start_bytecode)
                       if method.start_bytecode is not None else 0),
            "parameters": [
                self._construct_abi_type_hint(arg.type, arg_id) for arg_id, arg in method.args.items()
            ],
            "safe": method.is_safe,
        }

        return_type_extension = self._construct_abi_type_hint(method.type, is_return_type=True)
        for return_type_name, return_type in return_type_extension.items():
            method_abi[return_type_name] = return_type

        return method_abi

    @staticmethod
    def _construct_abi_type_hint(var_type: IType, var_id: str | None = None, is_return_type: bool = False) -> dict[str, Any] | None:
        """
        A recursive function that adds more details to some types on the manifest:
        - Arrays and Maps now have new keys to indicate the type of the items ('generic', 'generickey' and 'genericitem');
        - A String could have a 'hint' that it is an Address;
        - A Hash160 could have a 'hint' that it is a Scripthash or ScripthashLittleEndian;
        - A Hash256 could have a 'hint' that it is a BlockHash or TransactionId;
        - A StorageContext could have a 'hint' that it is a StorageContext or InteropInterface;
        - If the parameter or return is Optional, then the 'nullable' key will be added;
        - If the parameter or return is an Union, then the 'union' key will be added, with a list of types as value.
        """
        return_prefix = "return" if is_return_type else ""

        extended_type = {
            return_prefix + "type": var_type.abi_type
        }

        from boa3.internal.model.builtin.interop.interopinterfacetype import InteropInterfaceType
        from boa3.internal.model.type.annotation.uniontype import UnionType
        from boa3.internal.model.type.collection.sequence.mutable.listtype import ListType
        from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type
        from boa3.internal.model.type.collection.sequence.uint256type import UInt256Type
        from boa3.internal.model.type.collection.mapping.mutable.dicttype import DictType
        from boa3.internal.model.type.type import Type

        if isinstance(var_type, InteropInterfaceType):
            # Iterator or StorageContext is added as hint
            extended_type[return_prefix + "hint"] = var_type.raw_identifier

        # if it is a str, UInt160 or UInt256, then a type hint might be added
        elif any(
                isinstance(var_type, type(possible_type_hint)) and var_type.raw_identifier != possible_type_hint.raw_identifier
                for possible_type_hint in [Type.str, UInt160Type.build(), UInt256Type.build()]
        ):
            # Address, BlockHash, PublicKey, ScriptHash, ScriptHashLittleEndian or TransactionId is added as hint
            extended_type[return_prefix + "hint"] = var_type.raw_identifier

        # Calls itself to discover the types inside the Union/Optional
        elif isinstance(var_type, UnionType):

            from boa3.internal.model.type.annotation.optionaltype import OptionalType
            if isinstance(var_type, OptionalType):

                # if Optional is being used only with one type, e.g., Str | None, then don't consider it an Union
                if len(var_type.optional_types) == 1:
                    extended_type = FileGenerator._construct_abi_type_hint(var_type.optional_types[0],
                                                                           is_return_type=is_return_type)

                else:
                    extended_type[return_prefix + "union"] = [
                        FileGenerator._construct_abi_type_hint(union_type) for union_type in
                        var_type.optional_types
                    ]

                extended_type[return_prefix + "nullable"] = True

            else:
                extended_type[return_prefix + "union"] = [
                    FileGenerator._construct_abi_type_hint(union_type) for union_type in var_type.union_types
                ]

        # Calls itself to discover the types inside the List
        elif isinstance(var_type, ListType):
            extended_type[return_prefix + "generic"] = FileGenerator._construct_abi_type_hint(var_type.item_type)

        # Calls itself to discover the types inside the Dict
        elif isinstance(var_type, DictType):
            extended_type[return_prefix + "generickey"] = FileGenerator._construct_abi_type_hint(var_type.key_type)
            extended_type[return_prefix + "genericitem"] = FileGenerator._construct_abi_type_hint(var_type.item_type)

        if var_id is not None:
            extended_type[return_prefix + "name"] = var_id

        return extended_type

    def _get_abi_events(self) -> list[dict[str, Any]]:
        """
        Gets the abi events in a dictionary format

        :return: a dictionary with the abi events
        """
        return [
            {
                "name": name,
                "parameters": [
                    {
                        "name": arg_id,
                        "type": arg.type.abi_type
                    } for arg_id, arg in event.args_to_generate.items()
                ],
            } for name, event in self._events.items() if event.is_called
        ]

    def _get_extras(self) -> dict[str, Any] | None:
        """
        Gets the abi information in a dictionary format

        :return: a dictionary with the abi information
        """
        extras = {}
        for key, value in self._metadata.extras.items():
            try:
                json.dumps(value)  # include only json like extras
                extras[key] = value
            except BaseException:
                continue

        return extras if len(extras) > 0 else None

    # endregion

    # region Debug Info

    def generate_nefdbgnfo_file(self) -> bytes:
        """
        Generates a debug map for NEO debugger

        :return: the resulting map as a byte array
        """
        data: dict[str, Any] = self._get_debug_info()
        json_data: str = json.dumps(data, indent=4)
        return bytes(json_data, constants.ENCODING)

    def _get_debug_info(self) -> dict[str, Any]:
        """
        Gets the debug information in a dictionary format

        :return: a dictionary with the debug information
        """
        return {
            "hash": self._nef_hash,
            "entrypoint": self._entry_file_full_path,
            "documents": self._files,
            "static-variables": self._get_debug_static_variables(),
            "methods": self._get_debug_methods(),
            "events": self._get_debug_events()
        }

    def _get_debug_methods(self) -> list[dict[str, Any]]:
        """
        Gets the methods' debug information in a dictionary format

        :return: a dictionary with the methods' debug information
        """
        method_ids = []
        debug_methods = []

        for (module_id, method_id), method in self._methods_with_imports.items():
            dbg_method = self._get_method_debug_info(module_id, method_id, method)
            dbg_id = dbg_method['id']
            if method.is_compiled and dbg_id not in method_ids:
                method_ids.append(dbg_id)
                debug_methods.append(dbg_method)

        return debug_methods

    def _get_method_debug_info(self, module_id: str, method_id: str, method: Method) -> dict[str, Any]:
        from boa3.internal.compiler.codegenerator.vmcodemapping import VMCodeMapping
        from boa3.internal.neo.vm.type.AbiType import AbiType
        from boa3.internal.model.type.itype import IType

        sequence_points = []
        for instruction in method.debug_map():
            vm_code_map = VMCodeMapping.instance()
            start_address = vm_code_map.get_start_address(instruction.code)
            end_address = vm_code_map.get_end_address(instruction.code)
            if start_address >= method.start_address and end_address <= method.end_address:
                sequence_points.append(
                    '{0}[{1}]{2}:{3}-{4}:{5}'.format(start_address,
                                                     self._get_method_origin_index(method),
                                                     instruction.start_line, instruction.start_col,
                                                     instruction.end_line, instruction.end_col)
                )

        return {
            "id": str(id(method)),
            "name": '{0},{1}'.format(module_id, method_id),
            "range": '{0}-{1}'.format(method.start_address, method.end_address),
            "params": [
                '{0},{1}'.format(name, var.type.abi_type) for name, var in method.args.items()
            ],
            "return": method.return_type.abi_type,
            "variables": [
                '{0},{1}'.format(name, var.type.abi_type if isinstance(var.type, IType) else AbiType.Any)
                for name, var in method.locals.items()
            ],
            "sequence-points": sequence_points
        }

    def _get_method_origin_index(self, method: Method) -> int:
        if method.file_origin in self._files:
            return self._files.index(method.file_origin)

        imported_files: list[Import] = [imported for imported in self._symbols.values()
                                        if (isinstance(imported, Import)
                                            and not isinstance(imported, BuiltinImport)
                                            and imported.origin is not None)]
        imported = None
        for file in imported_files:
            if method in file.all_symbols.values():
                imported = file
                break

        if imported is None:
            return 0
        else:
            if imported.origin not in self._files:
                self._files.append(imported.origin)
            return self._files.index(imported.origin)

    def _get_debug_events(self) -> list[dict[str, Any]]:
        """
        Gets the events' debug information in a dictionary format

        :return: a dictionary with the event's debug information
        """
        event_ids = []
        debug_events = []

        for event_id, event in self._events.items():
            dbg_id = str(id(event))
            dbg_event = {
                "id": dbg_id,
                "name": ',{0}'.format(event_id),
                "params": [
                    '{0},{1}'.format(name, var.type.abi_type) for name, var in event.args.items()
                ]
            }
            if dbg_id not in event_ids:
                event_ids.append(dbg_id)
                debug_events.append(dbg_event)

        return debug_events

    def _get_debug_static_variables(self) -> list[str]:
        """
        Gets the static variables' debug information in a dictionary format

        :return: a dictionary with the event's debug information
        """
        from boa3.internal.model.type.itype import IType
        from boa3.internal.neo.vm.type.AbiType import AbiType

        static_variables = []

        for name, var in self._static_variables.items():
            var_unique_name = self._get_static_var_unique_name(name)
            var_type = var.type.abi_type if isinstance(var.type, IType) else AbiType.Any

            values = [var_unique_name, var_type.name]

            var_slot_index = self._get_static_var_slot_index(name)
            if isinstance(var_slot_index, int):
                values.append(str(var_slot_index))

            static_variables.append(
                constants.VARIABLE_NAME_SEPARATOR.join(values)
            )

        return static_variables

    # endregion

    def _get_static_var_unique_name(self, variable_id) -> str:
        imported_symbols: dict[str, Import] = {symbol.origin: symbol for symbol in self._all_imports}

        for name, symbol in self._symbols.items():
            if isinstance(symbol, Import) and symbol.origin not in imported_symbols:
                imported_symbols[symbol.origin] = symbol

        imported_to_map, imports_unique_ids = self._get_imports_unique_ids(imported_symbols,
                                                                           False,
                                                                           list(self._static_variables.values())
                                                                           )
        split_name = variable_id.split(constants.VARIABLE_NAME_SEPARATOR)
        if len(split_name) > 1:
            variable_original_id = split_name[-1]
            imported_id = constants.VARIABLE_NAME_SEPARATOR.join(split_name[:-1])
        else:
            variable_original_id = variable_id
            imported_id = None

        if isinstance(imported_id, str) and imported_id in imports_unique_ids:
            return '{0}.{1}'.format(imported_id, variable_original_id)

        if len(imported_to_map) <= 1:
            return variable_original_id

        for index, imported in enumerate(imported_symbols.values()):
            if isinstance(imported, Import) and variable_original_id in imported.all_symbols:
                if len(split_name) <= 1 or str(imported.ast.__hash__()) == split_name[0]:
                    return '{0}.{1}'.format(imports_unique_ids[index], variable_original_id)

        return '{0}.{1}'.format(imports_unique_ids[-1], variable_original_id)

    def _get_static_var_slot_index(self, variable_id) -> int | None:
        module_globals = list(self._static_variables.keys())
        if variable_id in module_globals:
            return module_globals.index(variable_id)
        return None

    def _get_imports_unique_ids(self, imported_symbols: dict[str, Import],
                                importing_methods: bool,
                                inner_imported_symbols: list[ISymbol] = None) -> tuple[dict[str, ImportData], list[str]]:
        if not isinstance(imported_symbols, dict):
            return {}, []
        if not isinstance(inner_imported_symbols, list):
            inner_imported_symbols = []

        # must map all imports, including inner imports
        index = 0
        while index < len(imported_symbols):
            module_origin, imported = list(imported_symbols.items())[index]
            for name, inner_imported in imported.all_symbols.items():
                if isinstance(inner_imported, Import) and inner_imported.origin not in inner_imported_symbols:
                    imported_symbols[inner_imported.origin] = inner_imported
            index += 1

        # map the modules that have user modules not imported by the entry file
        imported_to_map: dict[str, ImportData] = self._get_imports_to_map(imported_symbols, importing_methods)

        # change the full path names to unique small names
        imports_paths = list(imported_to_map)
        imports_paths.append(self._entry_file_full_path.replace('.py', '').replace('/__init__', ''))

        imports_unique_ids = []
        imports_duplicated_ids = []

        for index in range(len(imports_paths)):
            name = imports_paths[index]
            split_name = name.split(constants.PATH_SEPARATOR)
            index = -1
            short_name = split_name[index]
            while short_name in imports_duplicated_ids:
                index -= 1
                short_name = constants.ATTRIBUTE_NAME_SEPARATOR.join(split_name[index:])

            if short_name in imports_unique_ids:
                index_of_duplicated = imports_unique_ids.index(short_name)

                dup_split_name = imports_paths[index_of_duplicated].split(constants.PATH_SEPARATOR)
                dup_new_id = short_name
                dup_index = -len(imports_unique_ids[index].split(constants.ATTRIBUTE_NAME_SEPARATOR))

                while dup_new_id == short_name:
                    imports_duplicated_ids.append(short_name)
                    index -= 1
                    dup_index -= 1
                    short_name = constants.ATTRIBUTE_NAME_SEPARATOR.join(split_name[index:])
                    dup_new_id = constants.ATTRIBUTE_NAME_SEPARATOR.join(dup_split_name[dup_index:])

                imports_unique_ids[index_of_duplicated] = dup_new_id

            imports_unique_ids.append(short_name)

        return imported_to_map, imports_unique_ids

    def _get_imports_to_map(self, imported_symbols: dict[str, Import],
                            importing_methods: bool) -> dict[str, ImportData]:

        imported_to_map: dict[str, ImportData] = {}
        inner_packages: list[tuple[str, Package]] = []

        for name, imported in imported_symbols.items():
            need_to_map = False
            filtered_name = name.replace('.py', '').replace('/__init__', '')

            if not isinstance(imported, BuiltinImport):
                for _, symbol in imported.all_symbols.items():
                    if isinstance(symbol, Package):
                        inner_packages.append((filtered_name, symbol))

                    elif self._need_to_map_symbol(symbol, importing_methods):
                        need_to_map = True

            if need_to_map:
                if os.path.isfile(name):
                    origin_file = name
                else:
                    origin_file = None

                imported_to_map[filtered_name] = ImportData(imported_symbols[name], filtered_name, origin_file)
                if isinstance(origin_file, str) and origin_file not in self._files:
                    self._files.append(origin_file)

        while len(inner_packages) > 0:
            origin, package = inner_packages.pop(0)
            package_origin = f'{origin}/{package.raw_identifier}'
            for child in package.inner_packages.values():
                inner_packages.append((package_origin, child))

            if isinstance(package.origin, Import):
                need_to_map = self._need_to_map_symbol(package.origin, False)
                origin_file = package.origin.origin

                if not os.path.isfile(origin_file):
                    origin_file = None

                if isinstance(origin_file, str) and origin_file not in self._files:
                    self._files.append(origin_file)
            else:
                origin_file = None
                need_to_map = True

            if need_to_map:
                imported_to_map[package_origin] = ImportData(package, package_origin, origin_file)

        return imported_to_map

    def _need_to_map_symbol(self, symbol: ISymbol, importing_methods: bool) -> bool:
        from boa3.internal.model.builtin.builtincallable import IBuiltinCallable

        return (not importing_methods  # is importing variables or is a method but not builtin
                or (((isinstance(symbol, Method) and not isinstance(symbol, IBuiltinCallable))
                     or (isinstance(symbol, ClassType) and len(symbol.symbols) > 0)
                     ))
                )



from __future__ import annotations

__all__ = [
    'UInt160',
    'UInt256',
    'ECPoint',
    'Event',
    'Address',
    'BlockHash',
    'PublicKey',
    'ScriptHash',
    'ScriptHashLittleEndian',
    'TransactionId',
    'Block',
    "Signer",
    "WitnessRule",
    "WitnessCondition",
    "WitnessConditionType",
    "WitnessRuleAction",
    "WitnessScope",
    'Transaction',
    'Contract',
    'ContractManifest',
    'ContractPermission',
    'ContractPermissionDescriptor',
    'ContractGroup',
    'ContractAbi',
    'ContractMethodDescriptor',
    'ContractEventDescriptor',
    'ContractParameterDefinition',
    'ContractParameterType',
    'Nep17Contract',
    'Opcode',
    'FindOptions',
    'NamedCurveHash',
    'IBls12381',
    'Role',
    'Notification',
    'TriggerType',
    'VMState',
    'CallFlags',
    'OracleResponseCode',
    'NeoAccountState',
    'TransactionAttributeType',
]

from typing import Any

from boa3.internal.neo.vm.opcode.Opcode import Opcode
from boa3.internal.neo.vm.type.ContractParameterType import ContractParameterType
from boa3.internal.neo3.contracts import CallFlags
from boa3.internal.neo3.contracts import TriggerType
from boa3.internal.neo3.contracts.findoptions import FindOptions
from boa3.internal.neo3.contracts.namedcurvehash import NamedCurveHash
from boa3.internal.neo3.contracts.native import Role
from boa3.internal.neo3.network.payloads import OracleResponseCode
from boa3.internal.neo3.network.payloads.transactionattributetype import TransactionAttributeType
from boa3.internal.neo3.network.payloads.verification import WitnessScope, WitnessRuleAction, WitnessConditionType
from boa3.internal.neo3.vm import VMState


class UInt160(bytes):
    """
    Represents a 160-bit unsigned integer.
    """
    zero: UInt160

    def __init__(self, arg: bytes | int = 0):
        super().__init__()
        pass


class UInt256(bytes):
    """
    Represents a 256-bit unsigned integer.
    """
    zero: UInt256

    def __init__(self, arg: bytes | int = 0):
        super().__init__()
        pass


class ECPoint(bytes):
    """
    Represents a coordinate pair for elliptic curve cryptography (ECC) structures.
    """
    zero: ECPoint

    def __init__(self, arg: bytes):
        super().__init__()
        pass

    def to_script_hash(self) -> UInt160:
        """
        Converts a data to a script hash.

        :return: the script hash of the data
        :rtype: bytes
        """
        pass


class Event:
    """
    Describes an action that happened in the blockchain.
    Neo3-Boa compiler won't recognize the `__init__` of this class. To create a new Event, use the method `CreateNewEvent`:

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/develop/write/basics#events>`__ to learn more
    about Events.

    >>> from boa3.sc.utils import CreateNewEvent
    ... new_event: Event = CreateNewEvent(  # create a new Event with the CreateNewEvent method
    ...     [
    ...        ('name', str),
    ...        ('amount', int)
    ...     ],
    ...     'New Event'
    ... )
    """

    def __call__(self, *args, **kwargs):
        pass


Address = str
"""
A type used only to indicate that a parameter or return on the manifest should be treated as an Address.
Same as str.

:meta hide-value:
"""


BlockHash = UInt256
"""
A type used only to indicate that a parameter or return on the manifest should be treated as a BlockHash.
Same as UInt256.

:meta hide-value:
"""


PublicKey = ECPoint
"""
A type used only to indicate that a parameter or return on the manifest should be treated as a PublicKey.
Same as ECPoint.

:meta hide-value:
"""


ScriptHash = UInt160
"""
A type used only to indicate that a parameter or return on the manifest should be treated as a ScriptHash.
Same as UInt160.

:meta hide-value:
"""


ScriptHashLittleEndian = UInt160
"""
A type used only to indicate that a parameter or return on the manifest should be treated as a
ScriptHashLittleEndian.
Same as UInt160.

:meta hide-value:
"""


TransactionId = UInt256
"""
A type used only to indicate that a parameter or return on the manifest should be treated as a TransactionId.
Same as UInt256.

:meta hide-value:
"""


class Block:
    """
    Represents a block.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/foundation/Blocks>`__ to learn more
    about Blocks.

    :ivar hash: a unique identifier based on the unsigned data portion of the object
    :vartype hash: boa3.sc.types.UInt256
    :ivar version: the data structure version of the block
    :vartype version: int
    :ivar previous_hash: the hash of the previous block
    :vartype previous_hash: boa3.sc.types.UInt256
    :ivar merkle_root: the merkle root of the transactions
    :vartype merkle_root: boa3.sc.types.UInt256
    :ivar timestamp: UTC timestamp of the block in milliseconds
    :vartype timestamp: int
    :ivar nonce: a random number used once in the cryptography
    :vartype nonce: int
    :ivar index: the index of the block
    :vartype index: int
    :ivar primary_index: the primary index of the consensus node that generated this block
    :vartype primary_index: int
    :ivar next_consensus: the script hash of the consensus nodes that generates the next block
    :vartype next_consensus: boa3.sc.types.UInt160
    :ivar transaction_count: the number of transactions on this block
    :vartype transaction_count: int
    """

    def __init__(self):
        self.hash: UInt256 = UInt256()
        self.version: int = 0
        self.previous_hash: UInt256 = UInt256()
        self.merkle_root: UInt256 = UInt256()
        self.timestamp: int = 0
        self.nonce: int = 0
        self.index: int = 0
        self.primary_index: int = 0
        self.next_consensus: UInt160 = UInt160()
        self.transaction_count: int = 0


class Contract:
    """
    Represents a contract that can be invoked.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/glossary#smart-contract>`__ to learn about
    Smart Contracts.

    :ivar id: the serial number of the contract
    :vartype id: int
    :ivar update_counter: the number of times the contract was updated
    :vartype update_counter: int
    :ivar hash: the hash of the contract
    :vartype hash: boa3.sc.types.UInt160
    :ivar nef: the serialized Neo Executable Format (NEF) object holding of the smart contract code and compiler
        information
    :vartype nef: bytes
    :ivar manifest: the manifest of the contract
    :vartype manifest: ContractManifest
    """

    def __init__(self):
        self.id: int = 0
        self.update_counter: int = 0
        self.hash: UInt160 = UInt160()
        self.nef: bytes = bytes()
        self.manifest: ContractManifest = ContractManifest()


class ContractManifest:
    """
    Represents the manifest of a smart contract.

    When a smart contract is deployed, it must explicitly declare the features and permissions it will use.

    When it is running, it will be limited by its declared list of features and permissions, and cannot make any
    behavior beyond the scope of the list.

    For more details, check out `NEP-15 <https://github.com/neo-project/proposals/blob/master/nep-15.mediawiki>`__ or
    `Neo's Documentation <https://developers.neo.org/docs/n3/develop/write/manifest#manifest>`__.

    :ivar name: The name of the contract.
    :vartype name: str
    :ivar groups: The groups of the contract.
    :vartype groups: list[ContractGroup]
    :ivar supported_standards: Indicates which standards the contract supports. It can be a list of NEPs.
    :vartype supported_standards: list[str]
    :ivar abi: The ABI of the contract.
    :vartype abi: ContractAbi
    :ivar permissions: The permissions of the contract.
    :vartype permissions: list[ContractPermission]
    :ivar trusts:
        The trusted contracts and groups of the contract.

        If a contract is trusted, the user interface will not give any warnings when called by the contract.

    :vartype trusts: list[ContractPermissionDescriptor] or None
    :ivar extras: Custom user data as a json string.
    :vartype extras: str
    """

    def __init__(self):
        self.name: str = ''
        self.groups: list[ContractGroup] = []
        self.supported_standards: list[str] = []
        self.abi: ContractAbi = None
        self.permissions: list[ContractPermission] = []
        self.trusts: list[ContractPermissionDescriptor] | None = None
        self.extras: str = ''


class ContractPermission:
    """
    Represents a permission of a contract. It describes which contracts may be invoked and which methods are called.

    If a contract invokes a contract or method that is not declared in the manifest at runtime, the invocation will
    fail.

    :ivar contract:
        Indicates which contract to be invoked.

        It can be a hash of a contract, a public key of a group, or a wildcard \\*.

        If it specifies a hash of a contract, then the contract will be invoked; If it specifies a public key of a
        group, then any contract in this group may be invoked; If it specifies a wildcard \\*, then any contract may be
        invoked.

    :vartype contract: ContractPermissionDescriptor or None
    :ivar methods:
        Indicates which methods to be called.

        It can also be assigned with a wildcard \\*. If it is a wildcard \\*, then it means that any method can be
        called.

    :vartype methods: list[str] or None
    """

    def __init__(self):
        self.contract: ContractPermissionDescriptor | None = None
        self.methods: list[str] | None = None


class ContractPermissionDescriptor:
    """
    Indicates which contracts are authorized to be called.

    :ivar hash: The hash of the contract.
    :vartype hash: boa3.sc.types.UInt160 or None
    :ivar group: The group of the contracts.
    :vartype group: boa3.sc.types.ECPoint or None
    """

    def __init__(self):
        self.hash: UInt160 | None = None
        self.group: ECPoint | None = None


class ContractGroup:
    """
    Represents a set of mutually trusted contracts.

    A contract will trust and allow any contract in the same group to invoke it, and the user interface will not give
    any warnings.

    A group is identified by a public key and must be accompanied by a signature for the contract hash to prove that
    the contract is indeed included in the group.

    :ivar pubkey: The public key of the group.
    :vartype pubkey: boa3.sc.types.ECPoint
    :ivar signature: The signature of the contract hash which can be verified by `pubkey`.
    :vartype signature: bytes
    """

    def __init__(self):
        self.pubkey: ECPoint = ECPoint(b'')
        self.signature: bytes = b''


class ContractAbi:
    """
    Represents the ABI of a smart contract.

    For more details, check out `NEP-14 <https://github.com/neo-project/proposals/blob/master/nep-14.mediawiki>`__ or
    `Neo's Documentation <https://developers.neo.org/docs/n3/develop/write/manifest#manifest>`__.

    :ivar methods: Gets the methods in the ABI.
    :vartype methods: list[ContractMethodDescriptor]
    :ivar events: Gets the events in the ABI.
    :vartype events: list[ContractEventDescriptor]
    """

    def __init__(self):
        self.methods: list[ContractMethodDescriptor] = []
        self.events: list[ContractEventDescriptor] = []


class ContractMethodDescriptor:
    """
    Represents a method in a smart contract ABI.

    :ivar name: The name of the method.
    :vartype name: str
    :ivar parameters: The parameters of the method.
    :vartype parameters: list[ContractParameterDefinition]
    :ivar return_type: Indicates the return type of the method.
    :vartype return_type: ContractParameterType
    :ivar offset: The position of the method in the contract script.
    :vartype offset: int
    :ivar safe:
        Indicates whether the method is a safe method.

        If a method is marked as safe, the user interface will not give any warnings when it is called by other
        contracts.

    :vartype safe: bool
    """

    def __init__(self):
        self.name: str = ''
        self.parameters: list[ContractParameterDefinition] = []
        self.return_type: ContractParameterType = ContractParameterType.Any
        self.offset: int = 0
        self.safe: bool = False


class ContractEventDescriptor:
    """
    Represents an event in a smart contract ABI.

    :ivar name: The name of the event.
    :vartype name: str
    :ivar parameters: The parameters of the event.
    :vartype parameters: list[ContractParameterDefinition]
    """

    def __init__(self):
        self.name: str = ''
        self.parameters: list[ContractParameterDefinition] = []


class ContractParameterDefinition:
    """
    Represents a parameter of an event or method in ABI.

    :ivar name: The name of the parameter.
    :vartype name: str
    :ivar type: The type of the parameter.
    :vartype type: ContractParameterType
    """

    def __init__(self):
        self.name: str = ''
        self.type: ContractParameterType = ContractParameterType.Any


class Nep17Contract(Contract):
    def symbol(self) -> str:
        pass

    def decimals(self) -> int:
        pass

    def total_supply(self) -> int:
        pass

    def balance_of(self, account: UInt160) -> int:
        pass

    def transfer(self, from_address: UInt160, to_address: UInt160, amount: int, data: Any = None) -> bool:
        pass


class Signer:
    """
    Represents a signer.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/foundation/Transactions#signers>`__ to learn more
    about Signers.

    :ivar account:
    :vartype account: boa3.sc.types.UInt160
    :ivar scopes:
    :vartype scopes: WitnessScope
    :ivar allowed_contracts:
    :vartype allowed_contracts: list[boa3.sc.types.UInt160]
    :ivar allowed_groups:
    :vartype allowed_groups: list[boa3.sc.types.UInt160]
    :ivar rules:
    :vartype rules: list[WitnessRule]
    """

    def __init__(self):
        self.account: UInt160 = UInt160()
        self.scopes: WitnessScope = WitnessScope.NONE
        self.allowed_contracts: list[UInt160] = []
        self.allowed_groups: list[UInt160] = []
        self.rules: list[WitnessRule] = []


class WitnessRule:
    """
    Represents a witness rule.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/foundation/Transactions#witnessrule>`__ to learn
    more about WitnessRules.

    :ivar action:
    :vartype action: WitnessRuleAction
    :ivar condition:
    :vartype condition: WitnessCondition
    """

    def __init__(self):
        self.action: WitnessRuleAction = WitnessRuleAction.DENY
        self.condition: WitnessCondition = WitnessCondition()


class WitnessCondition:
    """
    Represents a witness condition.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/foundation/Transactions#witnesscondition>`__ to
    learn more about WitnessConditions.

    :ivar type:
    :vartype type: WitnessConditionType
    """

    def __init__(self):
        self.type: WitnessConditionType = WitnessConditionType.BOOLEAN


class Transaction:
    """
    Represents a transaction.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/foundation/Transactions>`__ to learn more about
    Transactions.

    :ivar hash: a unique identifier based on the unsigned data portion of the object
    :vartype hash: boa3.sc.types.UInt256
    :ivar version: the data structure version of the transaction
    :vartype version: int
    :ivar nonce: a random number used once in the cryptography
    :vartype nonce: int
    :ivar sender: the sender is the first signer of the transaction, they will pay the fees of the transaction
    :vartype sender: boa3.sc.types.UInt160
    :ivar system_fee: the fee paid for executing the `script`
    :vartype system_fee: int
    :ivar network_fee: the fee paid for the validation and inclusion of the transaction in a block by the consensus node
    :vartype network_fee: int
    :ivar valid_until_block: indicates that the transaction is only valid before this block height
    :vartype valid_until_block: int
    :ivar script: the array of instructions to be executed on the transaction chain by the virtual machine
    :vartype script: bytes
    """

    def __init__(self):
        self.hash: UInt256 = UInt256()
        self.version: int = 0
        self.nonce: int = 0
        self.sender: UInt160 = UInt160()
        self.system_fee: int = 0
        self.network_fee: int = 0
        self.valid_until_block: int = 0
        self.script: bytes = b''


class IBls12381:

    def __init__(self):
        pass


class Notification:
    """
    Represents a notification.

    :ivar script_hash: the script hash of the notification sender
    :vartype script_hash: boa3.sc.types.UInt160
    :ivar event_name: the notification's name
    :vartype event_name: str
    :ivar state: a tuple value storing all the notification contents.
    :vartype state: tuple
    """

    def __init__(self):
        self.script_hash: UInt160 = UInt160()
        self.event_name: str = ''
        self.state: tuple = ()


class NeoAccountState:
    """
    Represents the account state of NEO token in the NEO system.

    :ivar balance: the current account balance, which equals to the votes cast
    :vartype balance: int
    :ivar height: the height of the block where the balance changed last time
    :vartype height: int
    :ivar vote_to: the voting target of the account
    :vartype vote_to: boa3.sc.types.ECPoint
    """

    def __init__(self):
        from boa3.internal import constants
        self.balance: int = 0
        self.height: int = 0
        self.vote_to: ECPoint = ECPoint(bytes(constants.SIZE_OF_ECPOINT))
        self.last_gas_per_vote: int = 0



__all__ = [
    'ShowBlockCommand',
    'ShowTransactionCommand',
]

from boa3_test.test_drive.neoxp.command.neoexpresscommand.show.block import ShowBlockCommand
from boa3_test.test_drive.neoxp.command.neoexpresscommand.show.transaction import ShowTransactionCommand



from collections.abc import Sized
from typing import Any

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.collection.sequence.mutable.mutablesequencetype import MutableSequenceType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class AppendMethod(IBuiltinMethod):
    def __init__(self, sequence_type: MutableSequenceType = None):
        if not isinstance(sequence_type, MutableSequenceType):
            from boa3.internal.model.type.type import Type
            self_arg = Variable(Type.mutableSequence)
            item_arg = Variable(Type.any)
        else:
            self_arg = Variable(sequence_type)
            item_arg = Variable(sequence_type.value_type)

        identifier = 'append'
        args: dict[str, Variable] = {'self': self_arg, 'item': item_arg}
        super().__init__(identifier, args)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    @property
    def stores_on_slot(self) -> bool:
        return True

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 2:
            return False
        if not all(isinstance(param, IExpression) for param in params):
            return False

        from boa3.internal.model.type.itype import IType
        sequence_type: IType = params[0].type
        value_type: IType = params[1].type

        if not isinstance(sequence_type, MutableSequenceType):
            return False
        return sequence_type.value_type.is_type_of(value_type)

    def generate_opcodes(self, code_generator):
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.model.type.type import Type
        from boa3.internal.neo.vm.type.StackItem import StackItemType

        code_generator.duplicate_stack_item(2)
        code_generator.insert_type_check(StackItemType.Buffer)  # append opcode only works for array
        # if self is bytes
        if_bytes = code_generator.convert_begin_if()
        #   result = self + arg     # when it's bytearray, concatenates the value
        code_generator.convert_operation(BinaryOp.Concat.build(Type.bytearray, Type.bytearray))

        # else:
        else_bytes = code_generator.convert_begin_else(if_bytes, is_internal=True)
        #   self.append()
        super().generate_opcodes(code_generator)

        code_generator.convert_end_if(else_bytes, is_internal=True)

    def generate_internal_opcodes(self, code_generator):
        code_generator.insert_opcode(Opcode.APPEND)

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, Sized) and len(value) > 0:
            value = value[0]
        if value == self.args['self'].type:
            return self
        if isinstance(value, MutableSequenceType):
            return AppendMethod(value)
        return super().build(value)



from typing import Any, Self

from boa3.internal.neo.smart_contract.notification import Notification
from boa3.internal.neo3.core.types import UInt160
from boa3_test.test_drive.model.smart_contract.contractcollection import ContractCollection
from boa3_test.test_drive.model.smart_contract.testcontract import TestContract


class TestRunnerNotification(Notification):
    _script_hash_key = 'contract'
    _value_key = 'state'

    def __init__(self, event_name: str, script_hash: bytes, *value: Any):
        super().__init__(event_name, script_hash, *value)
        self._contract = None

    @property
    def contract(self) -> TestContract:
        return self._contract

    @classmethod
    def from_json(
            cls,
            json: dict[str, Any],
            contract_collection: ContractCollection = None,
            *args,
            **kwargs
    ) -> Self:

        result = super().from_json(json)
        if result is None:
            return result

        if isinstance(contract_collection, ContractCollection) and result.origin in contract_collection:
            result._contract = contract_collection[result.origin]
        return result

    @classmethod
    def _get_script_from_str(cls, script: str) -> bytes:
        if isinstance(script, str):
            if script.startswith('0x'):
                str_script = script[2:]
            else:
                str_script = script
            script = UInt160.from_string(str_script).to_array()

        return script



def deprecated(deprecated_in=None, removed_in=None, current_version=None, details=""):
    if current_version is None:
        import boa3
        current_version = boa3.__version__

    check_module = deprecated_in is None
    details = f'\n{details}'.replace('\n', '\n' + ' ' * 3)

    def wrapper(*args, **kwargs):
        deprecated_version = deprecated_in
        if check_module and len(args) > 0:
            import inspect
            import os.path

            if f'boa3{os.path.sep}builtin' in inspect.getfile(args[0]):
                import boa3
                deprecated_version = boa3.__boa3_builtin_deprecate_version__

        from deprecation import deprecated as deprecated_
        wrap = deprecated_(
            deprecated_in=deprecated_version,
            removed_in=removed_in,
            current_version=current_version,
            details=details
        )
        return wrap(*args, **kwargs)

    return wrapper



import ast

from boa3.internal.model.builtin.interop.nativecontract.Oracle.getoraclescripthashmethod import OracleContract
from boa3.internal.model.builtin.interop.nativecontract.nativecontractmethod import NativeContractMethod
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable


class OracleMethod(NativeContractMethod):

    def __init__(self, identifier: str, syscall: str, args: dict[str, Variable] = None,
                 defaults: list[ast.AST] = None, return_type: IType = None,
                 internal_call_args: int = None):
        super().__init__(OracleContract.getter, identifier, syscall, args, defaults, return_type, internal_call_args)



from typing import Any, Self

from boa3.internal.neo3.core.types import UInt160
from boa3_test.test_drive.model.wallet import utils
from boa3_test.test_drive.model.wallet.account import Account


class NeoExpressAccount(Account):
    def __init__(self, script_hash: UInt160,
                 account_version: int = 53,  # default Neo account version
                 name: str = None,
                 label: str = None):

        super().__init__(script_hash, account_version, name=name, label=label)

    @property
    def address(self) -> str | None:
        return utils.address_from_script_hash(self._script_hash.to_array(), self._version)

    @classmethod
    def from_json(cls, json: dict[str, Any]) -> Self:
        address = json['script-hash']
        label = json['label']
        account = NeoExpressAccount(
            script_hash=UInt160(utils.address_to_script_hash(address)),
            label=label
        )
        return account



from boa3.internal.model.builtin.interop.interop import Interop
from boa3.internal.model.builtin.native import *
from boa3.internal.model.builtin.native.oracleclass import OracleClass
from boa3.internal.model.identifiedsymbol import IdentifiedSymbol
from boa3.internal.model.imports.package import Package


class NativeContract:
    # Class Interfaces
    ContractManagement = ContractManagementClass.build()
    CryptoLib = CryptoLibClass.build()
    GAS = GasClass.build()
    Ledger = LedgerClass.build()
    NEO = NeoClass.build()
    Oracle = OracleClass.build()
    Policy = PolicyClass.build()
    RoleManagement = RoleManagementClass.build()
    StdLib = StdLibClass.build()

    # Deprecated Native Contract Names
    gas_name = 'GAS'
    ledger_name = 'Ledger'
    neo_name = 'NEO'
    oracle_name = 'Oracle'
    policy_name = 'Policy'

    # region Packages

    ContractManagementModule = Package(deprecated=True,
                                       new_location='boa3.sc.contracts',

                                       identifier=ContractManagement.identifier.lower(),
                                       types=[ContractManagement,
                                              Interop.ContractType])

    CryptoLibModule = Package(deprecated=True,
                              new_location='boa3.sc.contracts',
                              identifier=CryptoLib.identifier.lower(),
                              types=[CryptoLib,
                                     Interop.NamedCurveHashType,
                                     Interop.Bls12381Type])

    GasModule = Package(deprecated=True,
                        new_location='boa3.sc.contracts',
                        identifier=gas_name.lower(),
                        other_symbols={
                            gas_name: GAS
                        }
                        )

    LedgerModule = Package(deprecated=True,
                           new_location='boa3.sc.contracts',
                           identifier=ledger_name.lower(),
                           other_symbols={
                               ledger_name: Ledger
                           }
                           )

    NeoModule = Package(deprecated=True,
                        new_location='boa3.sc.contracts',
                        identifier=neo_name.lower(),
                        other_symbols={
                            neo_name: NEO
                        }
                        )

    OracleModule = Package(deprecated=True,
                           new_location='boa3.sc.contracts',
                           identifier=oracle_name.lower(),
                           other_symbols={
                               oracle_name: Oracle
                           }
                           )

    PolicyModule = Package(deprecated=True,
                           new_location='boa3.sc.contracts',
                           identifier=policy_name.lower(),
                           other_symbols={
                               policy_name: Policy
                           }
                           )

    RoleManagementModule = Package(deprecated=True,
                                   new_location='boa3.sc.contracts',
                                   identifier=RoleManagement.identifier.lower(),
                                   types=[RoleManagement,
                                          Interop.RoleType]
                                   )

    StdLibModule = Package(deprecated=True,
                           new_location='boa3.sc.contracts',
                           identifier=StdLib.identifier.lower(),
                           types=[StdLib]
                           )

    # endregion

    package_symbols: list[IdentifiedSymbol] = [
        ContractManagementModule,
        CryptoLibModule,
        GasModule,
        LedgerModule,
        NeoModule,
        OracleModule,
        PolicyModule,
        RoleManagementModule,
        StdLibModule
    ]



from boa3.internal.model.builtin.method.listmethod import ListMethod
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class ListMappingMethod(ListMethod):

    def __init__(self, mapping_type: IType = None):
        from boa3.internal.model.type.type import Type
        if mapping_type is None:
            mapping_type = Type.mapping

        args: dict[str, Variable] = {
            'value': Variable(mapping_type),
        }

        return_type = Type.list.build_collection(mapping_type.key_type)

        super().__init__(args, return_type)

    def generate_pack_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin

        code_generator.convert_builtin_method_call(Builtin.DictKeys, is_internal=True)
        code_generator.insert_opcode(Opcode.UNPACK)



from boa3.internal.model.builtin.builtinproperty import IBuiltinProperty
from boa3.internal.model.builtin.interop.contractgethashmethod import ContractGetHashMethod
from boa3.internal.model.variable import Variable


class GetLedgerScriptHashMethod(ContractGetHashMethod):
    def __init__(self):
        from boa3.internal.constants import LEDGER_SCRIPT
        from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type
        identifier = '-get_ledger_contract'
        args: dict[str, Variable] = {}
        super().__init__(LEDGER_SCRIPT, identifier, args, return_type=UInt160Type.build())

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return None


class LedgerContractProperty(IBuiltinProperty):
    def __init__(self):
        identifier = 'LedgerContract'
        getter = GetLedgerScriptHashMethod()
        super().__init__(identifier, getter)


LedgerContract = LedgerContractProperty()



from typing import Any

from boa3.internal.model.type.enum.intenumtype import IntEnumType
from boa3.internal.model.type.itype import IType
from boa3.internal.neo3.network.payloads import OracleResponseCode


class OracleResponseCodeType(IntEnumType):
    """
    A class used to represent Neo's OracleResponseCode.
    """

    def __init__(self):
        super().__init__(OracleResponseCode)
        self._identifier = 'OracleResponseCode'

    @classmethod
    def build(cls, value: Any = None) -> IType:
        if value is None or cls._is_type_of(value):
            return _OracleResponseCode

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, (OracleResponseCode, OracleResponseCodeType))


_OracleResponseCode = OracleResponseCodeType()



from boa3.internal.model.builtin.method import IBuiltinMethod
from boa3.internal.model.variable import Variable


class StorageMapGetMethod(IBuiltinMethod):

    def __init__(self):
        from boa3.internal.model.builtin.interop.storage.storagemap.storagemaptype import _StorageMap
        from boa3.internal.model.type.type import Type

        identifier = 'get'
        args: dict[str, Variable] = {'self': Variable(_StorageMap),
                                     'key': Variable(Type.bytes)}

        super().__init__(identifier, args, return_type=Type.bytes)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def self_type(self):
        """
        :rtype: boa3.internal.model.builtin.interop.storage.storagemap.storagemaptype.StorageMapType
        """
        return self.args['self'].type

    @property
    def _body(self) -> str | None:
        return None

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.interop.interop import Interop
        from boa3.internal.model.operation.binaryop import BinaryOp

        # actual_key = self._prefix + key
        code_generator.swap_reverse_stack_items(2)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_load_class_variable(self.self_type, '_prefix', is_internal=True)  # self._prefix

        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_operation(BinaryOp.Concat, is_internal=True)

        # actual_context = self._context
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_load_class_variable(self.self_type, '_context', is_internal=True)  # self._context

        # return get(actual_key, actual_context)
        code_generator.convert_builtin_method_call(Interop.StorageGet, is_internal=True)



import logging
import os
import platform
import stat
import sys

import requests

logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)


def get_umask():
    umask = os.umask(0)
    os.umask(umask)
    return umask


def chmod_plus_x(path):
    os.chmod(
        path,
        os.stat(path).st_mode |
        (
            (
                stat.S_IXUSR |
                stat.S_IXGRP |
                stat.S_IXOTH
            )
            & ~get_umask()
        )
    )


def main():
    if not sys.maxsize > 2 ** 32:
        print("CPM is not supported on 32-bit platforms")

    r = requests.get("https://api.github.com/repos/CityOfZion/cpm/releases")

    for release in r.json():
        tag = release['tag_name']
        version = tag[1:]
        if platform.machine().lower() in ["x86_64", "amd64"]:
            arch = "amd64"
        else:
            arch = "arm64"
        system = platform.system().lower()
        asset_needle = f"cpm_{version}_{system}_{arch}"
        if system == "windows":
            asset_needle += ".exe"

        for asset in release['assets']:
            if asset['name'] == asset_needle:
                r = requests.get(asset['browser_download_url'], stream=True)
                print(f"Found release! Downloading {asset['browser_download_url']}...", end='')

                cpm_filename = f"{os.path.dirname(sys.argv[0])}/cpm"
                if system == "windows":
                    cpm_filename += ".exe"

                with open(cpm_filename, 'wb') as f:
                    for chunk in r.iter_content(chunk_size=1024):
                        if chunk:  # filter out keep-alive new chunks
                            f.write(chunk)
                if system in ["darwin", "linux"]:
                    chmod_plus_x(cpm_filename)
                print("done")
                return


if __name__ == "__main__":
    main()



__all__ = [
    "Iterator",
    "env",
    "CreateNewEvent",
    "Nep11TransferEvent",
    "Nep17TransferEvent",
    "to_hex_str",
    "to_script_hash",
    "call_contract",
    "get_call_flags",
    "abort",
    "check_sig",
    "check_multisig",
    "create_standard_account",
    "create_multisig_account",
    "to_bool",
    "to_int",
    "to_bytes",
    "to_str",
    "hash160",
    "hash256",
]

from typing import Any, Sequence

from boa3.sc.types import Event, ECPoint, UInt160, CallFlags
from boa3.sc.utils.iterator import Iterator

env: str = 'compiler'
"""
Gets the compiled environment. This allows specific environment validations to be easily included in the smart contract 
logic without the need to rewrite anything before compiling (i.e. changes in smart contracts hashes between testnet and 
mainnet). 

>>> # compiling with 'neo3-boa compile -e test_net ./path/to/contract.py'
... from boa3.sc.utils import call_contract
... from boa3.sc.types import UInt160
... call_contract(UInt160(b'12345678901234567890') if env == 'test_net' else b'abcdeabcdeabcdeabcde',
...               'balanceOf',
...               UInt160(b'zyxwvzyxwvzyxwvzyxwv'))
110000

>>> # compiling with 'neo3-boa compile -e main_net ./path/to/contract.py'
... from boa3.sc.utils import call_contract
... from boa3.sc.types import UInt160
... call_contract(UInt160(b'12345678901234567890') if env == 'test_net' else b'abcdeabcdeabcdeabcde',
...               'balanceOf',
...               UInt160(b'zyxwvzyxwvzyxwvzyxwv'))
250

:meta hide-value:
"""


def CreateNewEvent(arguments: list[tuple[str, type]] = [], event_name: str = '') -> Event:
    """
    Creates a new Event.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/develop/write/basics#events>`__ to learn more
    about Events.

    >>> new_event: Event = CreateNewEvent(
    ...     [
    ...        ('name', str),
    ...        ('amount', int)
    ...     ],
    ...     'New Event'
    ... )

    :param arguments: the list of the events args' names and types
    :type arguments: list[tuple[str, type]]
    :param event_name: custom name of the event. It's filled with the variable name if not specified
    :type event_name: str
    :return: the new event
    :rtype: Event
    """
    pass


Nep11TransferEvent: Event = CreateNewEvent(
    [
        ('from', UInt160 | None),
        ('to', UInt160 | None),
        ('amount', int),
        ('tokenId', str | bytes)
    ],
    'Transfer'
)
"""
The NEP-11 Transfer event that should be triggered whenever a non-fungible token is transferred, minted or burned. It 
needs the addresses of the sender, receiver, amount transferred and the id of the token.

Check out the `proposal <https://github.com/neo-project/proposals/blob/master/nep-11.mediawiki>`__ or 
`Neo's Documentation <https://developers.neo.org/docs/n3/develop/write/nep11>`__ about this NEP.

>>> Nep11TransferEvent(b'\\xd1\\x17\\x92\\x82\\x12\\xc6\\xbe\\xfa\\x05\\xa0\\x23\\x07\\xa1\\x12\\x55\\x41\\x06\\x55\\x10\\xe6',  # when calling, it will return None, but the event will be triggered
...                    b'\\x18\\xb7\\x30\\x14\\xdf\\xcb\\xee\\x01\\x30\\x00\\x13\\x9b\\x8d\\xa0\\x13\\xfb\\x96\\xac\\xd1\\xc0', 1, '01')
{
    'name': 'Transfer',
    'script hash': b'\\x13\\xb4\\x51\\xa2\\x1c\\x10\\x12\\xd6\\x13\\x12\\x19\\x0c\\x15\\x61\\x9b\\x1b\\xd1\\xa2\\xf4\\xb2',
    'state': {
        'from': b'\\xd1\\x17\\x92\\x82\\x12\\xc6\\xbe\\xfa\\x05\\xa0\\x23\\x07\\xa1\\x12\\x55\\x41\\x06\\x55\\x10\\xe6',
        'to': b'\\x18\\xb7\\x30\\x14\\xdf\\xcb\\xee\\x01\\x30\\x00\\x13\\x9b\\x8d\\xa0\\x13\\xfb\\x96\\xac\\xd1\\xc0',
        'amount': 1,
        'tokenId': '01'
    }
}

:meta hide-value:
"""


Nep17TransferEvent: Event = CreateNewEvent(
    [
        ('from', UInt160 | None),
        ('to', UInt160 | None),
        ('amount', int)
    ],
    'Transfer'
)
"""
The NEP-17 Transfer event that should be triggered whenever a fungible token is transferred, minted or burned. It needs
the addresses of the sender, receiver and the amount transferred.

Check out the `proposal <https://github.com/neo-project/proposals/blob/master/nep-17.mediawiki>`__ or 
`Neo's Documentation <https://developers.neo.org/docs/n3/develop/write/nep17>`__ about this NEP.

>>> Nep17TransferEvent(b'\\xd1\\x17\\x92\\x82\\x12\\xc6\\xbe\\xfa\\x05\\xa0\\x23\\x07\\xa1\\x12\\x55\\x41\\x06\\x55\\x10\\xe6',  # when calling, it will return None, but the event will be triggered
...                    b'\\x18\\xb7\\x30\\x14\\xdf\\xcb\\xee\\x01\\x30\\x00\\x13\\x9b\\x8d\\xa0\\x13\\xfb\\x96\\xac\\xd1\\xc0', 100)
{
    'name': 'Transfer',
    'script hash': b'\\x17\\xe3\\xca\\x91\\xca\\xb7\\xaf\\xdd\\xe6\\xba\\x07\\xaa\\xba\\xa1\\x66\\xab\\xcf\\x00\\x04\\x50',
    'state': {
        'from': b'\\xd1\\x17\\x92\\x82\\x12\\xc6\\xbe\\xfa\\x05\\xa0\\x23\\x07\\xa1\\x12\\x55\\x41\\x06\\x55\\x10\\xe6',
        'to': b'\\x18\\xb7\\x30\\x14\\xdf\\xcb\\xee\\x01\\x30\\x00\\x13\\x9b\\x8d\\xa0\\x13\\xfb\\x96\\xac\\xd1\\xc0',
        'amount': 100
    }
}

:meta hide-value:
"""


def to_hex_str(data: bytes) -> str:
    """
    Converts bytes into its string hex representation.

    >>> to_hex_str(ECPoint(bytes(range(33))))
    '201f1e1d1c1b1a191817161514131211100f0e0d0c0b0a09080706050403020100'

    >>> to_hex_str(b'1234567891')
    '31393837363534333231'

    :param data: data to represent as hex.
    :type data: bytearray or bytes

    :return: the hex representation of the data
    :rtype: str
    """
    pass


def to_script_hash(data_bytes: Any) -> bytes:
    """
    Converts a data to a script hash.

    >>> to_script_hash(ECPoint(bytes(range(33))))
    b'\\x12\\xc8z\\xfb3k\\x1e4>\\xb3\\x83\\tK\\xc7\\xdch\\xe5\\xee\\xc7\\x98'

    >>> to_script_hash(b'1234567891')
    b'\\x4b\\x56\\x34\\x17\\xed\\x99\\x7f\\x13\\x22\\x67\\x40\\x79\\x36\\x8b\\xa2\\xcd\\x72\\x41\\x25\\x6d'

    :param data_bytes: data to hash
    :type data_bytes: Any
    :return: the script hash of the data
    :rtype: bytes
    """
    pass


def call_contract(script_hash: UInt160, method: str, args: Sequence = (), call_flags: CallFlags = CallFlags.ALL) -> Any:
    """
    Calls a smart contract given the method and the arguments. Since the return is type Any, you'll probably need to
    type cast the return.

    >>> from boa3.sc.contracts import NeoToken
    ... call_contract(NeoToken.hash, 'balanceOf', [UInt160(b'\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2')])
    100

    :param script_hash: the target smart contract's script hash
    :type script_hash: boa3.sc.types.UInt160
    :param method: the name of the method to be executed
    :type method: str
    :param args: the specified method's arguments
    :type args: Sequence[Any]
    :param call_flags: the CallFlags to be used to call the contract
    :type call_flags: boa3.sc.types.CallFlags

    :return: the result of the specified method
    :rtype: Any

    :raise Exception: raised if there isn't a valid CallFlags, the script hash is not a valid smart contract or the
        method was not found or the arguments aren't valid to the specified method.
    """
    pass


def get_call_flags() -> CallFlags:
    """
    Gets the CallFlags in the current context.

    >>> get_call_flags()
    CallFlags.READ_ONLY
    """
    pass


def abort(msg: str | None = None):
    """
    Aborts the execution of a smart contract. Using this will cancel the changes made on the blockchain by the
    transaction.

    >>> abort()     # abort doesn't return anything by itself, but the execution will stop and the VMState will be FAULT
    VMState.FAULT

    >>> abort('abort message')
    VMState.FAULT

    """
    pass


def check_sig(pub_key: ECPoint, signature: bytes) -> bool:
    """
    Checks the signature for the current script container.

    >>> check_sig(ECPoint(b'\\x03\\x5a\\x92\\x8f\\x20\\x16\\x39\\x20\\x4e\\x06\\xb4\\x36\\x8b\\x1a\\x93\\x36\\x54\\x62\\xa8\\xeb\\xbf\\xf0\\xb8\\x81\\x81\\x51\\xb7\\x4f\\xaa\\xb3\\xa2\\xb6\\x1a'),
    ...           b'wrongsignature')
    False

    :param pub_key: the public key of the account
    :type pub_key: boa3.sc.types.ECPoint
    :param signature: the signature of the current script container
    :type signature: bytes
    :return: whether the signature is valid or not
    :rtype: bool
    """
    pass


def check_multisig(pubkeys: list[ECPoint], signatures: list[bytes]) -> bool:
    """
    Checks the signatures for the current script container.

    >>> check_multisig([ECPoint(b"\\x03\\xcd\\xb0\\x67\\xd9\\x30\\xfd\\x5a\\xda\\xa6\\xc6\\x85\\x45\\x01\\x60\\x44\\xaa\\xdd\\xec\\x64\\xba\\x39\\xe5\\x48\\x25\\x0e\\xae\\xa5\\x51\\x17\\x2e\\x53\\x5c"),
    ...                 ECPoint(b"\\x03l\\x841\\xccx\\xb31w\\xa6\\x0bK\\xcc\\x02\\xba\\xf6\\r\\x05\\xfe\\xe5\\x03\\x8es9\\xd3\\xa6\\x88\\xe3\\x94\\xc2\\xcb\\xd8C")],
    ...                [b'wrongsignature1', b'wrongsignature2'])
    False

    :param pubkeys: a list of public keys
    :type pubkeys: list[boa3.sc.types.ECPoint]
    :param signatures: a list of signatures
    :type signatures: list[bytes]
    :return: a boolean value that represents whether the signatures were validated
    :rtype: bool
    """
    pass


def create_standard_account(pub_key: ECPoint) -> UInt160:
    """
    Calculates the script hash from a public key.

    >>> create_standard_account(ECPoint(b'\\x03\\x5a\\x92\\x8f\\x20\\x16\\x39\\x20\\x4e\\x06\\xb4\\x36\\x8b\\x1a\\x93\\x36\\x54\\x62\\xa8\\xeb\\xbf\\xf0\\xb8\\x81\\x81\\x51\\xb7\\x4f\\xaa\\xb3\\xa2\\xb6\\x1a'))
    b'\\r\\xa9g\\xa4\\x00C+\\xf2\\x7f\\x8e\\x8e\\xb4o\\xe8\\xace\\x9e\\xcc\\xde\\x04'

    :param pub_key: the given public key
    :type pub_key: boa3.sc.types.ECPoint

    :return: the corresponding script hash of the public key
    :rtype: boa3.sc.types.UInt160
    """
    pass


def create_multisig_account(m: int, pub_keys: list[ECPoint]) -> UInt160:
    """
    Calculates corresponding multisig account script hash for the given public keys.

    >>> create_multisig_account(1, [ECPoint(b'\\x03\\x5a\\x92\\x8f\\x20\\x16\\x39\\x20\\x4e\\x06\\xb4\\x36\\x8b\\x1a\\x93\\x36\\x54\\x62\\xa8\\xeb\\xbf\\xf0\\xb8\\x81\\x81\\x51\\xb7\\x4f\\xaa\\xb3\\xa2\\xb6\\x1a')])
    b'"5,\\xd2\\x9e\\xe7\\xb4\\x02\\x08b\\xdbd\\x1e\\xedx\\x82\\x8fU(m'

    :param m: the minimum number of correct signatures need to be provided in order for the verification to pass.
    :type m: int
    :param pub_keys: the public keys of the account
    :type pub_keys: list[boa3.sc.types.ECPoint]

    :return: the hash of the corresponding account
    :rtype: boa3.sc.types.UInt160
    """
    pass


def to_bool(value: bytes) -> bool:
    """
    Return a bytes value to the boolean it represents.

    >>> to_bool(b'\\x00')
    False

    >>> to_bool(b'\\x01')
    True

    >>> to_bool(b'\\x02')
    True
    """
    pass


def to_bytes(value: str | int, length: int = None, big_endian: bool = False, signed: bool = True) -> bytes:
    """
    Converts a str or integer value to an array of bytes.
    If the value is a string, the other parameters are not valid.
    If the value is an integer, the length of the bytes and whether it's expecting big or little endian representation
    can be specified.
    Throws an exception if the int value is negative or if the length is less than the length of the value in bytes.

    >>> to_bytes(65)
    b'A'
    >>> to_bytes(1234)
    b'\xd2\x04'
    >>> to_bytes(-1)
    b'\xff'

    >>> to_bytes(1234, 3)
    b'\xd2\x04\x00'
    >>> to_bytes(1234, 3, True)
    b'\x00\x04\xd2'
    >>> to_bytes(1234, 3, False)
    b'\xd2\x04\x00'
    >>> to_bytes(-1234, 3, False)
    b'\x2e\xfb\xff'

    >>> to_bytes(-120, 1, False, True)
    b'\x88'

    >>> to_bytes('A')
    b'A'

    :param value: value to be converted to bytes
    :type value: str | int
    :param length: available only to integer values, it represents the length of the resulting bytes. If value is None,
    then it will try to use the least amount of bytes possible
    :type length: int
    :param big_endian: available only to integer values, whether to represent the integer in big-endian (True) or
    little-endian (False) byte order
    :type big_endian: bool
    :param signed: available only to integer values, whether to represent the integer as signed (True) or
    unsigned (False) in bytes return
    :type signed: bool
    :return: a byte value that represents the given value
    :rtype: bytes

    :raise Exception: raised if int value is negative or length is less than the length that the integer value can be represented
    """
    pass


def to_int(value: bytes, big_endian: bool = False, signed: bool = True) -> int:
    """
    Converts a bytes value to the integer it represents.

    >>> to_int(b'A')
    65

    >>> to_int(b'\xfa\x15')
    5626

    >>> to_int(b'\xfa\x15', False)
    5626
    >>> to_int(b'\xfa\x15', True)
    -1515

    >>> to_int(b'\xfa\x15', True, False)
    64021
    >>> to_int(b'\xfa\x15', False, False)
    5626

    :param value: value to be converted to int
    :type value: bytes
    :param big_endian: whether to interpret the bytes in big-endian (True) or little-endian (False) byte order
    :type big_endian: bool
    :param signed: whether to interpret the bytes as a signed (True) or unsigned (False) integer
    :type signed: bool
    :return:
    :rtype: int
    """
    pass


def to_str(value: bytes) -> str:
    """
    Converts a bytes value to a string.

    >>> to_str(b'A')
    'A'
    """
    pass


def hash160(key: Any) -> bytes:
    """
    Encrypts a key using HASH160.

    >>> hash160('unit test')
    b'#Q\\xc9\\xaf+c\\x12\\xb1\\xb9\\x9e\\xa1\\x89t\\xa228g\\xec\\x0eF'

    >>> hash160(10)
    b'\\x89\\x86D\\x19\\xa8\\xc3v%\\x00\\xfe\\x9a\\x98\\xaf\\x8f\\xbbO3u\\x08\\xf0'

    :param key: the key to be encrypted
    :type key: Any
    :return: a byte value that represents the encrypted key
    :rtype: bytes
    """
    pass


def hash256(key: Any) -> bytes:
    """
    Encrypts a key using HASH256.

    >>> hash256('unit test')
    b'\\xdau1>J\\xc2W\\xf8LN\\xfb2\\x0f\\xbd\\x01\\x1cr@<\\xf5\\x93<\\x90\\xd2\\xe3\\xb8$\\xd6H\\x96\\xf8\\x9a'

    >>> hash256(10)
    b'\\x9c\\x82r\\x01\\xb9@\\x19\\xb4/\\x85pk\\xc4\\x9cY\\xff\\x84\\xb5`M\\x11\\xca\\xaf\\xb9\\n\\xb9HV\\xc4\\xe1\\xddz'

    :param key: the key to be encrypted
    :type key: Any
    :return: a byte value that represents the encrypted key
    :rtype: bytes
    """
    pass



__all__ = [
    'StorageContext',
    'StorageMap',
    'get',
    'get_int',
    'get_bool',
    'get_str',
    'get_list',
    'get_dict',
    'get_object',
    'get_uint160',
    'get_uint256',
    'get_ecpoint',
    'try_get',
    'try_get_int',
    'try_get_bool',
    'try_get_str',
    'try_get_list',
    'try_get_dict',
    'try_get_object',
    'try_get_uint160',
    'try_get_uint256',
    'try_get_ecpoint',
    'get_context',
    'get_read_only_context',
    'put',
    'put_int',
    'put_bool',
    'put_str',
    'put_list',
    'put_dict',
    'put_object',
    'put_uint160',
    'put_uint256',
    'put_ecpoint',
    'delete',
    'find',
]

from typing import Any

from boa3.sc.storage.storagecontext import StorageContext
from boa3.sc.storage.storagemap import StorageMap
from boa3.sc.types import FindOptions, UInt256, ECPoint, UInt160
from boa3.sc.utils.iterator import Iterator


def get_context() -> StorageContext:
    """
    Gets current storage context.

    >>> get_context()       # StorageContext cannot be read outside the blockchain
    _InteropInterface

    :return: the current storage context
    :rtype: StorageContext
    """
    pass


def get(key: bytes, context: StorageContext = get_context()) -> bytes:
    """
    Gets a value from the persistent store based on the given key.

    >>> put(b'unit', b'test')
    ... get(b'unit')
    b'test'

    >>> get(b'fake_key')
    b''

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context
    :rtype: bytes
    """
    pass


def try_get(key: bytes, context: StorageContext = get_context()) -> tuple[bytes, bool]:
    """
    Gets a value from the persistent store based on the given key and returns whether the value is stored.

    >>> put(b'unit', b'test')
    ... try_get(b'unit')
    (b'test', True)

    >>> try_get(b'fake_key')
    (b'', False)

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context and whether it was actually stored
    :rtype: tuple[bytes, bool]
    """
    pass


def get_int(key: bytes, context: StorageContext = get_context()) -> int:
    """
    Gets a value as integer from the persistent store based on the given key.
    It's equivalent to boa3.sc.utils.to_int(get(key, context))

    >>> put_int(b'unit', 5)
    ... get_int(b'unit')
    5

    >>> get_int(b'fake_key')
    0

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context
    :rtype: int
    """
    pass


def try_get_int(key: bytes, context: StorageContext = get_context()) -> tuple[int, bool]:
    """
    Gets a value as integer from the persistent store based on the given key and returns whether the value is stored.

    >>> put_int(b'unit', 5)
    ... try_get_int(b'unit')
    (5, True)

    >>> try_get_int(b'fake_key')
    (0, False)

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context and whether it was actually stored
    :rtype: tuple[int, bool]
    """
    pass


def get_bool(key: bytes, context: StorageContext = get_context()) -> bool:
    """
    Gets a value as boolean from the persistent store based on the given key.
    It's equivalent to boa3.sc.utils.to_bool(get(key, context))

    >>> put_bool(b'unit', True)
    ... get_bool(b'unit')
    True

    >>> get_bool(b'fake_key')
    False

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context
    :rtype: bool
    """
    pass


def try_get_bool(key: bytes, context: StorageContext = get_context()) -> tuple[bool, bool]:
    """
    Gets a value as boolean from the persistent store based on the given key and returns whether the value is stored.

    >>> put_bool(b'unit', False)
    ... try_get_bool(b'unit')
    (False, True)

    >>> try_get_bool(b'fake_key')
    (False, False)

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context and whether it was actually stored
    :rtype: tuple[bool, bool]
    """
    pass


def get_str(key: bytes, context: StorageContext = get_context()) -> str:
    """
    Gets a value as string from the persistent store based on the given key.
    It's equivalent to boa3.sc.utils.to_str(get(key, context))

    >>> put_str(b'unit', 'test')
    ... get_str(b'unit')
    'test'

    >>> get_str(b'fake_key')
    ''

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context
    :rtype: str
    """
    pass


def try_get_str(key: bytes, context: StorageContext = get_context()) -> tuple[str, bool]:
    """
    Gets a value as string from the persistent store based on the given key and returns whether the value is stored.

    >>> put_str(b'unit', 'test')
    ... try_get_str(b'unit')
    ('test', True)

    >>> try_get_str(b'fake_key')
    ('', False)

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context and whether it was actually stored
    :rtype: tuple[str, bool]
    """
    pass


def get_list(key: bytes, context: StorageContext = get_context()) -> list:
    """
    Gets a value as list from the persistent store based on the given key.
    It's equivalent to boa3.sc.contracts.stdlib.StdLib.deserialize(get(key, context))

    >>> put_list(b'unit', [1, 2, '3'])
    ... get_list(b'unit')
    [1, 2, '3']

    >>> get_list(b'fake_key')
    []

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context
    :rtype: list
    """
    pass


def try_get_list(key: bytes, context: StorageContext = get_context()) -> tuple[list, bool]:
    """
    Gets a value as list from the persistent store based on the given key and returns whether the value is stored.

    >>> put_list(b'unit', [1, 2, '3'])
    ... try_get_list(b'unit')
    ([1, 2, '3'], True)

    >>> get_list(b'fake_key')
    ([], False)

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context and whether it was actually stored
    :rtype: tuple[list, bool]
    """
    pass


def get_dict(key: bytes, context: StorageContext = get_context()) -> dict:
    """
    Gets a value as dict from the persistent store based on the given key.
    It's equivalent to boa3.sc.contracts.stdlib.StdLib.deserialize(get(key, context))

    >>> put_dict(b'unit', {'example': 1, 'other_example': 2})
    ... get_dict(b'unit')
    {'example': 1, 'other_example': 2}

    >>> get_dict(b'fake_key')
    {}

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context
    :rtype: dict
    """
    pass


def try_get_dict(key: bytes, context: StorageContext = get_context()) -> tuple[dict, bool]:
    """
    Gets a value as dict from the persistent store based on the given key and returns whether the value is stored.

    >>> put_dict(b'unit', {'example': 1, 'other_example': 2})
    ... get_dict(b'unit')
    ({'example': 1, 'other_example': 2}, True)

    >>> get_dict(b'fake_key')
    ({}, False)

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context and whether it was actually stored
    :rtype: tuple[dict, bool]
    """
    pass


def get_object(key: bytes, context: StorageContext = get_context()) -> Any:
    """
    Gets a value as object from the persistent store based on the given key.
    It's equivalent to boa3.sc.contracts.stdlib.StdLib.deserialize(get(key, context))
    Returns an empty struct if not found.

    >>> example = SomeClass()
    >>> put_object(b'unit', example)
    ... get_object(b'unit')
    SomeClass

    >>> get_object(b'fake_key')
    object

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context
    :rtype: Any
    """
    pass


def try_get_object(key: bytes, context: StorageContext = get_context()) -> tuple[Any, bool]:
    """
    Gets a value as dict from the persistent store based on the given key and returns whether the value is stored.

    >>> example = SomeClass()
    >>> put_object(b'unit', example)
    ... try_get_object(b'unit')
    (SomeClass, True)

    >>> get_object(b'fake_key')
    (object, False)

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context and whether it was actually stored
    :rtype: tuple[Any, bool]
    """
    pass


def get_uint160(key: bytes, context: StorageContext = get_context()) -> UInt160:
    """
    Gets a value as UInt160 from the persistent store based on the given key.
    It's equivalent UInt160(get(key, context))

    >>> put_uint160(b'unit', UInt160(b'0123456789ABCDEFGHIJ'))
    ... get_uint160(b'unit')
    UInt160(0x4a49484746454443424139383736353433323130)

    >>> get_uint160(b'fake_key')
    UInt160(0x0000000000000000000000000000000000000000)

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context
    :rtype: boa3.sc.types.UInt160
    """
    pass


def try_get_uint160(key: bytes, context: StorageContext = get_context()) -> tuple[UInt160, bool]:
    """
    Gets a value as UInt160 from the persistent store based on the given key and returns whether the value is stored.

    >>> put_uint160(b'unit', UInt160(b'0123456789ABCDEFGHIJ'))
    ... try_get_uint160(b'unit')
    (UInt160(0x4a49484746454443424139383736353433323130), True)

    >>> get_uint160(b'fake_key')
    (UInt160(0x0000000000000000000000000000000000000000), False)

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context and whether it was actually stored
    :rtype: tuple[boa3.sc.types.UInt160, bool]
    """
    pass


def get_uint256(key: bytes, context: StorageContext = get_context()) -> UInt256:
    """
    Gets a value as UInt256 from the persistent store based on the given key.
    It's equivalent UInt256(get(key, context))

    >>> put_uint256(b'unit', UInt256(b'0123456789ABCDEFGHIJKLMNOPQRSTUV'))
    ... get_uint256(b'unit')
    UInt256(0x565554535251504f4e4d4c4b4a49484746454443424139383736353433323130)

    >>> get_uint160(b'fake_key')
    UInt256(0x0000000000000000000000000000000000000000000000000000000000000000)

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context
    :rtype: boa3.sc.types.UInt256
    """
    pass


def try_get_uint256(key: bytes, context: StorageContext = get_context()) -> tuple[UInt256, bool]:
    """
    Gets a value as UInt256 from the persistent store based on the given key and returns whether the value is stored.

    >>> put_uint256(b'unit', UInt256(b'0123456789ABCDEFGHIJKLMNOPQRSTUV'))
    ... get_uint256(b'unit')
    (UInt256(0x565554535251504f4e4d4c4b4a49484746454443424139383736353433323130), True)

    >>> get_uint160(b'fake_key')
    (UInt256(0x0000000000000000000000000000000000000000000000000000000000000000), False)

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context and whether it was actually stored
    :rtype: tuple[boa3.sc.types.UInt256, bool]
    """
    pass


def get_ecpoint(key: bytes, context: StorageContext = get_context()) -> ECPoint:
    """
    Gets a value as ECPoint from the persistent store based on the given key.
    It's equivalent ECPoint(get(key, context))

    >>> put_ecpoint(b'unit', ECPoint(b'0123456789ABCDEFGHIJKLMNOPQRSTUVW'))
    ... get_ecpoint(b'unit')
    ECPoint(0x57565554535251504f4e4d4c4b4a49484746454443424139383736353433323130)

    >>> get_ecpoint(b'fake_key')
    ECPoint(0x000000000000000000000000000000000000000000000000000000000000000000)

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context
    :rtype: boa3.sc.types.ECPoint
    """
    pass


def try_get_ecpoint(key: bytes, context: StorageContext = get_context()) -> tuple[ECPoint, bool]:
    """
    Gets a value as ECPoint from the persistent store based on the given key and returns whether the value is stored.

    >>> put_ecpoint(b'unit', ECPoint(b'0123456789ABCDEFGHIJKLMNOPQRSTUVW'))
    ... try_get_ecpoint(b'unit')
    (ECPoint(0x57565554535251504f4e4d4c4b4a49484746454443424139383736353433323130), True)

    >>> try_get_ecpoint(b'fake_key')
    (ECPoint(0x000000000000000000000000000000000000000000000000000000000000000000), False)

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context and whether it was actually stored
    :rtype: tuple[boa3.sc.types.ECPoint, bool]
    """
    pass


def get_read_only_context() -> StorageContext:
    """
    Gets current read only storage context.

    >>> get_context()       # StorageContext cannot be read outside the blockchain
    _InteropInterface

    :return: the current read only storage context
    :rtype: StorageContext
    """
    pass


def put(key: bytes, value: bytes, context: StorageContext = get_context()):
    """
    Inserts a given bytes value in the key-value format into the persistent storage.

    >>> put(b'unit', b'test')
    None

    :param key: the identifier in the store for the new value
    :type key: bytes
    :param value: value to be stored
    :type value: bytes
    :param context: storage context to be used
    :type context: StorageContext
    """
    pass


def put_int(key: bytes, value: int, context: StorageContext = get_context()):
    """
    Inserts a given integer value in the key-value format into the persistent storage.
    It's equivalent to put(key, boa3.sc.utils.to_int(value), context)

    >>> put_int(b'unit', 5)
    None

    :param key: the identifier in the store for the new value
    :type key: bytes
    :param value: value to be stored
    :type value: int
    :param context: storage context to be used
    :type context: StorageContext
    """
    pass


def put_bool(key: bytes, value: bool, context: StorageContext = get_context()):
    """
    Inserts a given boolean value in the key-value format into the persistent storage.
    It's equivalent to put(key, boa3.sc.utils.to_bool(value), context)

    >>> put_bool(b'unit', True)
    None

    :param key: the identifier in the store for the new value
    :type key: bytes
    :param value: value to be stored
    :type value: bool
    :param context: storage context to be used
    :type context: StorageContext
    """
    pass


def put_str(key: bytes, value: str, context: StorageContext = get_context()):
    """
    Inserts a given str value in the key-value format into the persistent storage.
    It's equivalent to put(key, boa3.sc.utils.to_str(value), context)

    >>> put_str(b'unit', 'test')
    None

    :param key: the identifier in the store for the new value
    :type key: bytes
    :param value: value to be stored
    :type value: str
    :param context: storage context to be used
    :type context: StorageContext
    """
    pass


def put_list(key: bytes, value: list, context: StorageContext = get_context()):
    """
    Inserts a given list value in the key-value format into the persistent storage.
    It's equivalent to put(key, boa3.sc.contracts.stdlib.StdLib.serialize(value), context)

    >>> put_list(b'unit', [1, 2, '3'])
    None

    :param key: the identifier in the store for the new value
    :type key: bytes
    :param value: value to be stored
    :type value: list
    :param context: storage context to be used
    :type context: StorageContext
    """
    pass


def put_dict(key: bytes, value: dict, context: StorageContext = get_context()):
    """
    Inserts a given dict value in the key-value format into the persistent storage.
    It's equivalent to put(key, boa3.sc.contracts.stdlib.StdLib.serialize(value), context)

    >>> put_dict(b'unit', {'example': 1, 'other_example': 2})
    None

    :param key: the identifier in the store for the new value
    :type key: bytes
    :param value: value to be stored
    :type value: dict
    :param context: storage context to be used
    :type context: StorageContext
    """
    pass


def put_object(key: bytes, value: object, context: StorageContext = get_context()):
    """
    Inserts a given object value in the key-value format into the persistent storage.
    It's equivalent to put(key, boa3.sc.contracts.stdlib.StdLib.serialize(value), context)

    >>> example = SomeClass()
    >>> put_object(b'unit', example)
    None

    :param key: the identifier in the store for the new value
    :type key: bytes
    :param value: value to be stored
    :type value: dict
    :param context: storage context to be used
    :type context: StorageContext
    """
    pass


def put_uint160(key: bytes, value: UInt160, context: StorageContext = get_context()):
    """
    Inserts a given UInt160 value in the key-value format into the persistent storage.
    It's equivalent to put(key, value, context) since UInt160 is a subclass of bytes

    >>> put_uint160(b'unit', UInt160(b'0123456789ABCDEFGHIJ'))
    None

    :param key: the identifier in the store for the new value
    :type key: bytes
    :param value: value to be stored
    :type value: boa3.sc.types.UInt160
    :param context: storage context to be used
    :type context: StorageContext
    """
    pass


def put_uint256(key: bytes, value: UInt256, context: StorageContext = get_context()):
    """
    Inserts a given UInt256 value in the key-value format into the persistent storage.
    It's equivalent to put(key, value, context) since UInt256 is a subclass of bytes

    >>> put_uint256(b'unit', UInt256(b'0123456789ABCDEFGHIJKLMNOPQRSTUV'))
    None

    :param key: the identifier in the store for the new value
    :type key: bytes
    :param value: value to be stored
    :type value: boa3.sc.types.UInt256
    :param context: storage context to be used
    :type context: StorageContext
    """
    pass


def put_ecpoint(key: bytes, value: ECPoint, context: StorageContext = get_context()):
    """
    Inserts a given ECPoint value in the key-value format into the persistent storage.
    It's equivalent to put(key, value, context) since ECPoint is a subclass of bytes

    >>> put_ecpoint(b'unit', ECPoint(b'0123456789ABCDEFGHIJKLMNOPQRSTUVW'))
    None

    :param key: the identifier in the store for the new value
    :type key: bytes
    :param value: value to be stored
    :type value: boa3.sc.types.ECPoint
    :param context: storage context to be used
    :type context: StorageContext
    """
    pass


def delete(key: bytes, context: StorageContext = get_context()):
    """
    Removes a given key from the persistent storage if exists.

    >>> put(b'unit', 'test')
    ... delete()
    ... get(b'unit')
    ''

    :param key: the identifier in the store for the new value
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    """
    pass


def find(prefix: bytes,
         context: StorageContext = get_context(),
         options: FindOptions = FindOptions.NONE) -> Iterator:
    """
    Searches in the storage for keys that start with the given prefix.

    >>> put(b'a1', 'one')
    ... put(b'a2', 'two')
    ... put(b'a3', 'three')
    ... put(b'b4', 'four')
    ... findIterator = find(b'a')
    ... findResults = []
    ... while findIterator.next():
    ...     findResults.append(findIterator.value)
    ... findResults
    ['one', 'two', 'three']

    :param prefix: prefix to find the storage keys
    :type prefix: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :param options: the options of the search
    :type options: boa3.sc.types.FindOptions
    :return: an iterator with the search results
    :rtype: Iterator
    """
    pass



import ast
from typing import Any

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.collection.sequence.sequencetype import SequenceType
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class ByteArrayMethod(IBuiltinMethod):

    def __init__(self, argument_type: IType | dict = None):
        identifier = 'bytearray'
        from boa3.internal.model.type.type import Type

        if isinstance(argument_type, dict):
            args = argument_type
            defaults = []
        else:
            from boa3.internal.model.type.type import Type
            if argument_type is None or not self.validate_parameters(argument_type):
                argument_type = Type.none

            args: dict[str, Variable] = {'object': Variable(argument_type)}
            object_default = ast.parse(f"{Type.int.default_value}"
                                       ).body[0].value
            defaults = [object_default]

        super().__init__(identifier, args, defaults=defaults, return_type=Type.bytearray)

    @property
    def _arg_object(self) -> Variable:
        return self.args['object']

    @property
    def identifier(self) -> str:
        from boa3.internal.model.type.type import Type
        if self._arg_object.type is Type.none:
            return self._identifier

        if self._args_on_stack == 2:
            return '-{0}_with_enconding'.format(self._identifier)

        return '-{0}_from_{1}'.format(self._identifier, self._arg_object.type._identifier)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) > 1:
            return False
        if len(params) == 0:
            return True

        from boa3.internal.model.type.itype import IType
        if not isinstance(params[0], (IExpression, IType)):
            return False

        param_type: IType = params[0].type if isinstance(params[0], IExpression) else params[0]
        from boa3.internal.model.type.type import Type

        if Type.int.is_type_of(param_type):
            return True

        return (isinstance(param_type, SequenceType)
                and (param_type is Type.str
                     or isinstance(param_type.value_type, type(Type.int))
                     ))

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.type.type import Type

        if Type.int.is_type_of(self._arg_object.type):
            code_generator.insert_opcode(Opcode.NEWBUFFER)
        else:
            code_generator.convert_cast(self.return_type, is_internal=True)

    @property
    def is_supported(self) -> bool:
        # TODO: change when building bytearray from int iterators are implemented #2tz74k9
        from boa3.internal.model.type.type import Type
        return self._arg_object.type in (Type.bytes, Type.str, Type.int)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self._arg_object.type):
            return self
        if isinstance(value, list):
            if len(value) == 2:
                from boa3.internal.model.builtin.method.bytearrayencodingmethod import ByteArrayEncodingMethod
                return ByteArrayEncodingMethod()

            value = value[0] if len(value) > 0 else None
        if self.validate_parameters(value):
            return ByteArrayMethod(value)
        return super().build(value)



__all__ = [
    'Bls12381AddMethod',
    'Bls12381DeserializeMethod',
    'Bls12381EqualMethod',
    'Bls12381MulMethod',
    'Bls12381PairingMethod',
    'Bls12381SerializeMethod',
    'Bls12381Type',
    'CheckMultisigMethod',
    'CheckSigMethod',
    'Hash160Method',
    'Hash256Method',
    'Keccak256Method',
    'Murmur32Method',
    'NamedCurveHashType',
    'Ripemd160Method',
    'Sha256Method',
    'VerifyWithECDsaMethod',
]

from boa3.internal.model.builtin.interop.crypto.bls12381addmethod import Bls12381AddMethod
from boa3.internal.model.builtin.interop.crypto.bls12381deserializemethod import Bls12381DeserializeMethod
from boa3.internal.model.builtin.interop.crypto.bls12381equalmethod import Bls12381EqualMethod
from boa3.internal.model.builtin.interop.crypto.bls12381mulmethod import Bls12381MulMethod
from boa3.internal.model.builtin.interop.crypto.bls12381pairingmethod import Bls12381PairingMethod
from boa3.internal.model.builtin.interop.crypto.bls12381serializemethod import Bls12381SerializeMethod
from boa3.internal.model.builtin.interop.crypto.bls12381type import Bls12381Type
from boa3.internal.model.builtin.interop.crypto.checkmultisigmethod import CheckMultisigMethod
from boa3.internal.model.builtin.interop.crypto.checksigmethod import CheckSigMethod
from boa3.internal.model.builtin.interop.crypto.hash160method import Hash160Method
from boa3.internal.model.builtin.interop.crypto.hash256method import Hash256Method
from boa3.internal.model.builtin.interop.crypto.keccak256method import Keccak256Method
from boa3.internal.model.builtin.interop.crypto.murmur32method import Murmur32Method
from boa3.internal.model.builtin.interop.crypto.namedcurvehashtype import NamedCurveHashType
from boa3.internal.model.builtin.interop.crypto.ripemd160method import Ripemd160Method
from boa3.internal.model.builtin.interop.crypto.sha256method import Sha256Method
from boa3.internal.model.builtin.interop.crypto.verifywithecdsa import VerifyWithECDsaMethod



from boa3.internal.model.builtin.interop.contract import ContractType
from boa3.internal.model.method import Method


class Nep17ContractClass(ContractType):
    def __init__(self):
        super().__init__()
        
        self._identifier = 'Nep17Contract'
        
        from boa3.internal.model.builtin.contract.nep17_interface_methods import (
            Nep17InterfaceBalanceOfMethod,
            Nep17InterfaceDecimalsMethod,
            Nep17InterfaceSymbolMethod,
            Nep17InterfaceTotalSupplyMethod,
            Nep17InterfaceTransferMethod,
        )

        nep17_methods: list = [
            Nep17InterfaceBalanceOfMethod,
            Nep17InterfaceDecimalsMethod,
            Nep17InterfaceSymbolMethod,
            Nep17InterfaceTotalSupplyMethod,
            Nep17InterfaceTransferMethod,
        ]

        instance_methods: dict[str, Method] = {}
        for nep17_method in nep17_methods:
            nep17_method_obj = nep17_method(self)
            instance_methods[nep17_method_obj.identifier] = nep17_method_obj

        self._instance_methods: dict[str, Method] = instance_methods

    @property
    def instance_methods(self) -> dict[str, Method]:
        return self._instance_methods.copy()



import os

from boa3.internal import constants

PROJECT_ROOT_DIRECTORY = constants.PATH_SEPARATOR.join(os.path.dirname(__file__).split(os.sep)[:-2])

# If you don't want to use the neo-express instance on boa3_test/tests on a lot of test, then change this path to
# another one, otherwise you could just pass the neoxp_path parameter when instantiating NeoTestRunner
NEO_EXPRESS_INSTANCE_DIRECTORY = f'{PROJECT_ROOT_DIRECTORY}/boa3_test/tests'
TEST_RUNNER_DIRECTORY = os.path.abspath('.')



from typing import Any, Self

from boa3.internal.model.method import Method
from boa3.internal.model.property import Property
from boa3.internal.model.type.classes.classstructtype import ClassStructType
from boa3.internal.model.variable import Variable


class ContractGroupType(ClassStructType):
    """
    A class used to represent Neo ContractGroup class
    """

    def __init__(self):
        super().__init__('ContractGroup')
        from boa3.internal.model.type.collection.sequence.ecpointtype import ECPointType
        from boa3.internal.model.type.type import Type

        self._variables: dict[str, Variable] = {
            'pubkey': Variable(ECPointType.build()),
            'signature': Variable(Type.bytes)
        }
        self._constructor: Method = None

    @property
    def class_variables(self) -> dict[str, Variable]:
        return {}

    @property
    def instance_variables(self) -> dict[str, Variable]:
        return self._variables.copy()

    @property
    def properties(self) -> dict[str, Property]:
        return {}

    @property
    def static_methods(self) -> dict[str, Method]:
        return {}

    @property
    def class_methods(self) -> dict[str, Method]:
        return {}

    @property
    def instance_methods(self) -> dict[str, Method]:
        return {}

    def constructor_method(self) -> Method | None:
        return self._constructor

    @classmethod
    def build(cls, value: Any = None) -> Self:
        if value is None or cls._is_type_of(value):
            return _ContractGroup

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, ContractGroupType)


_ContractGroup = ContractGroupType()



from boa3.internal.model.builtin.interop.storage.getcheck.istoragetrygetmethod import IStorageTryGetMethod


class StorageTryGetUInt256Method(IStorageTryGetMethod):
    def __init__(self):
        from boa3.internal.model.type.collection.sequence.uint256type import UInt256Type

        identifier = 'try_get_uint256'
        value_type = UInt256Type.build()

        super().__init__(identifier, value_type=value_type)

    def generate_default_value_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin

        code_generator.convert_literal(Builtin.UInt256.default_value)

    def generate_deserialize_value_opcodes(self, code_generator):
        # it doesn't need to deserialize bytes values
        return



from boa3.internal.model.builtin.method.printmethod import PrintMethod
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class PrintBoolMethod(PrintMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        arg_value = Type.bool

        super().__init__(arg_value)

    def _generate_print_opcodes(self, code_generator):
        from boa3.internal.model.builtin.interop.interop import Interop

        code_generator.insert_opcode(Opcode.NZ)
        code_generator.convert_builtin_method_call(Interop.JsonSerialize, is_internal=True)



__all__ = ['StringSplitMethod',
           'StrLenMethod',
           ]

from boa3.internal.model.builtin.native.stdlib.stringsplitmethod import StringSplitMethod
from boa3.internal.model.builtin.native.stdlib.strlenmethod import StrLenMethod



import ast
from typing import Any

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.type.primitive.ibytestringtype import IByteStringType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class ReplaceMethod(IBuiltinMethod):
    def __init__(self, self_type: IByteStringType = None):
        from boa3.internal.model.type.type import Type

        if not isinstance(self_type, IByteStringType):
            self_type = Type.bytes

        identifier = 'replace'
        args: dict[str, Variable] = {
            'self': Variable(self_type),
            'old': Variable(self_type),
            'new': Variable(self_type),
            'count': Variable(Type.int),
        }

        count_default = ast.parse("-1").body[0].value.operand
        count_default.n = -1

        super().__init__(identifier, args, defaults=[count_default], return_type=self_type)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.model.type.type import Type

        # old_length = len(old)
        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)

        # self_length = len(string)
        code_generator.duplicate_stack_item(5)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)

        # self_length = self_length - old_length + 1
        code_generator.duplicate_stack_item(2)
        code_generator.convert_operation(BinaryOp.Sub, is_internal=True)
        code_generator.insert_opcode(Opcode.INC)

        # if count < 0:
        code_generator.move_stack_item_to_top(5)
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(0)
        if_less_than_zero = code_generator.convert_begin_if()
        code_generator.change_jump(if_less_than_zero, Opcode.JMPGE)

        #     count = self_length
        code_generator.remove_stack_top_item()
        code_generator.duplicate_stack_top_item()

        code_generator.convert_end_if(if_less_than_zero)

        # index = 0
        code_generator.convert_literal(0)

        # new_string = ""
        code_generator.convert_literal(Type.str.default_value)

        # while index < self_length:
        while_start = code_generator.convert_begin_while()

        #   if self[index:index + old_length] == old and count > 0:
        #   (self[index:index + old_length] == old)
        code_generator.duplicate_stack_item(8)
        code_generator.duplicate_stack_item(3)
        code_generator.duplicate_stack_item(7)
        code_generator.convert_get_substring(is_internal=True)
        code_generator.duplicate_stack_item(7)


        if_substring_matches_old = code_generator.convert_begin_if()
        code_generator.change_jump(if_substring_matches_old, Opcode.JMPNE)

        #   (count > 0)
        code_generator.duplicate_stack_item(3)
        code_generator.convert_literal(0)

        if_count_greater_than_zero = code_generator.convert_begin_if()
        code_generator.change_jump(if_count_greater_than_zero, Opcode.JMPLE)

        #       new_string += new
        code_generator.duplicate_stack_item(7)
        code_generator.convert_operation(BinaryOp.Concat, is_internal=True)

        #       index += old_length
        code_generator.swap_reverse_stack_items(2)
        code_generator.duplicate_stack_item(5)
        code_generator.convert_operation(BinaryOp.Add, is_internal=True)
        code_generator.swap_reverse_stack_items(2)

        #       count -= 1
        code_generator.swap_reverse_stack_items(3)
        code_generator.insert_opcode(Opcode.DEC)
        code_generator.swap_reverse_stack_items(3)

        #   else:
        else_substring_matches_old = code_generator.convert_begin_else(if_substring_matches_old)
        code_generator.convert_end_if(if_count_greater_than_zero)

        #       new_string += self[index]
        code_generator.duplicate_stack_item(8)
        code_generator.duplicate_stack_item(3)
        code_generator.convert_literal(1)
        code_generator.convert_get_substring(is_internal=True)
        code_generator.convert_operation(BinaryOp.Concat, is_internal=True)

        #       index += 1
        code_generator.swap_reverse_stack_items(2)
        code_generator.insert_opcode(Opcode.INC)
        code_generator.swap_reverse_stack_items(2)

        code_generator.convert_end_if(else_substring_matches_old)

        # end while
        while_condition_start = code_generator.bytecode_size
        # index < self_length
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(5)
        code_generator.convert_operation(BinaryOp.Lt, is_internal=True)

        code_generator.convert_end_while(while_start, while_condition_start, is_internal=True)

        code_generator.remove_stack_item(2)
        code_generator.remove_stack_item(2)
        code_generator.remove_stack_item(2)

        code_generator.remove_stack_item(3)
        code_generator.remove_stack_item(3)

        # if old_length > 1:
        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(1)
        if_old_length_greater_than_one = code_generator.convert_begin_if()
        code_generator.change_jump(if_old_length_greater_than_one, Opcode.JMPLE)

        #     new_string += self[-(old_length - 1):]
        code_generator.swap_reverse_stack_items(3)
        code_generator.swap_reverse_stack_items(2)
        code_generator.insert_opcode(Opcode.DEC)
        code_generator.insert_opcode(Opcode.RIGHT, pop_from_stack=True, add_to_stack=[Type.str])
        code_generator.convert_operation(BinaryOp.Concat, is_internal=True)

        else_old_length_greater_than_one = code_generator.convert_begin_else(if_old_length_greater_than_one, is_internal=True)

        code_generator.remove_stack_item(2)
        code_generator.remove_stack_item(2)

        code_generator.convert_end_if(else_old_length_greater_than_one)

        # return new_string
        code_generator.convert_cast(Type.str, is_internal=True)

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, IByteStringType):
            return ReplaceMethod(value)
        return super().build(value)



import ast
from typing import Any, Self

from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.itype import IType


class Variable(IExpression):
    """
    A class used to represent a variable

    :ivar var_type: the type of the variable.
    """

    def __init__(self,
                 var_type: IType | None,
                 origin_node: ast.AST | None = None,
                 deprecated: bool = False
                 ):
        super().__init__(origin_node, deprecated)
        self.defined_by_entry = True

        from boa3.internal.analyser.model.optimizer import Undefined, UndefinedType
        if var_type is None:
            var_type = Undefined

        self._var_type: IType | UndefinedType = var_type

        self.is_reassigned = False
        self._origin_variable: Variable | None = None
        self._first_assign_value: Any = Undefined

    def copy(self) -> Self:
        var = Variable(self._var_type, self._origin_node)
        var.is_reassigned = self.is_reassigned
        var._first_assign_value = self._first_assign_value
        var._origin_variable = self._origin_variable if self._origin_variable is not None else self
        return var

    @property
    def shadowing_name(self) -> str:
        return 'variable'

    @property
    def type(self) -> IType:
        return self._var_type

    @property
    def is_global(self) -> bool:
        return hasattr(self, '_set_as_global') and self._set_as_global

    @property
    def has_literal_value(self) -> bool:
        from boa3.internal.analyser.model.optimizer import Undefined
        return self._first_assign_value is not Undefined

    def __repr__(self) -> str:
        return '{0}({1})'.format(self.__class__.__name__, self.__str__())

    def __str__(self) -> str:
        return str(self.type)

    def set_type(self, var_type: IType):
        """
        Sets a type for the variable

        :param var_type:
        """
        self._var_type = var_type

    def set_is_reassigned(self):
        if not self.is_reassigned:
            self.is_reassigned = True
        if hasattr(self._origin_variable, 'set_is_reassigned'):
            self._origin_variable.set_is_reassigned()

    def set_initial_assign(self, first_value: Any):
        if not self.has_literal_value:
            self._first_assign_value = first_value



__all__ = [
    'NeoEngine',
    'NeoEngineState'
]

from typing import Any

from boa3.internal.compiler.codegenerator.engine.executionscript import ExecutionScript
from boa3.internal.compiler.codegenerator.engine.istack import IStack
from boa3.internal.neo.vm.VMCode import VMCode
from boa3.internal.neo.vm.opcode.Opcode import Opcode
from boa3.internal.neo.vm.type.StackItem import StackItemType
from boa3.internal.neo3.vm import VMState


class StackItem:
    def __init__(self, value: Any, stack_item_type: StackItemType):
        self.value = value
        self.type = stack_item_type


class NeoStack(IStack):
    def __init__(self):
        super().__init__(stack_type=StackItem)

    def _default_constructor_args(self) -> tuple:
        return tuple()


class NeoEngineState:
    def __init__(self,
                 entry_point: int,
                 ):

        self.entry_point: int = entry_point
        self.current_instruction: int = entry_point
        self.result_stack: list[IStack] = []
        self.vm_state: VMState = VMState.NONE


class NeoEngine:
    def __init__(self):
        self._script: ExecutionScript = None
        self._state_stack: list[NeoEngineState] = []

    @property
    def current_state(self) -> NeoEngineState | None:
        if self._state_stack:
            return self._state_stack[-1]
        else:
            return None

    def load_script(self, script: ExecutionScript):
        self._script = script

    def execute(self, entry_point: int) -> set[int]:
        executed_instructions_addresses = set()

        if not isinstance(self._script, ExecutionScript):
            raise ValueError('Script was not loaded')

        if entry_point not in self._script:
            raise ValueError('Invalid address given as entry point')

        cur_address = entry_point
        self._state_stack.append(NeoEngineState(cur_address))
        while self._state_stack and cur_address in self._script:
            self._execute_instruction(self._script.get_instruction(cur_address))

            executed_instructions_addresses.add(cur_address)
            cur_address = self._script.next_address(cur_address)

        if self._state_stack:
            self._state_stack.clear()

        return executed_instructions_addresses

    def _execute_instruction(self, current_opcode: VMCode):
        match current_opcode.opcode:
            case Opcode.INITSLOT | Opcode.INITSSLOT:
                return
            case Opcode.RET:
                self._state_stack.pop()
